"""
a tkinter frame
"""
import os
import tkinter as tk
from tkinter import ttk
from time import time
from threading import Thread

from mmg_toolbox.utils.env_functions import get_dls_visits
from mmg_toolbox.beamline_metadata.config import BEAMLINE_CONFIG
from mmg_toolbox.utils.file_functions import folder_summary_line
from ..misc.logging import create_logger
from ..misc.config import get_config, save_config, C
from ..misc.functions import select_folder, show_error

logger = create_logger(__file__)


class TitleWindow:
    def __init__(self, root: tk.Misc, config: dict | None = None):
        t0 = time()
        self.root = root
        self.config = config or get_config()

        self.beamline = tk.StringVar(self.root, '')
        self.visit = tk.StringVar(self.root, '')
        self.summary = tk.StringVar(self.root, '')
        self.data_dir = tk.StringVar(self.root, '')
        self.proc_dir = tk.StringVar(self.root, '')
        self.notebook_dir = tk.StringVar(self.root, '')
        self.visits = {}

        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES)
        ttk.Label(frm, textvariable=self.beamline, style="title.TLabel").pack(side=tk.RIGHT)

        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES)
        ttk.Label(frm, text='Visit:').pack(side=tk.LEFT, padx=4)
        visits = list(self.visits.keys())
        first = next(iter(visits), '')
        self.visit_menu = ttk.OptionMenu(frm, self.visit, first,*visits, command=self.choose_visit)
        self.visit_menu.pack(side=tk.LEFT, padx=4)
        ttk.Button(frm, text='Check', command=self.open_file_browser, width=10).pack(side=tk.LEFT)
        ttk.Label(frm, textvariable=self.summary).pack(side=tk.LEFT, padx=2)

        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        ttk.Label(frm, text='Data Dir:', width=15).pack(side=tk.LEFT, padx=4)
        ttk.Entry(frm, textvariable=self.data_dir, width=60).pack(side=tk.LEFT)
        ttk.Button(frm, text='Browse', command=self.browse_datadir).pack(side=tk.LEFT)

        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        ttk.Label(frm, text='Analysis Dir:', width=15).pack(side=tk.LEFT, padx=4)
        ttk.Entry(frm, textvariable=self.proc_dir, width=60).pack(side=tk.LEFT)
        ttk.Button(frm, text='Browse', command=self.browse_analysis).pack(side=tk.LEFT)

        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        ttk.Label(frm, text='Notebook Dir:', width=15).pack(side=tk.LEFT, padx=4)
        ttk.Entry(frm, textvariable=self.notebook_dir, width=60).pack(side=tk.LEFT)
        ttk.Button(frm, text='Browse', command=self.browse_notebook).pack(side=tk.LEFT)

        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, pady=6)
        ttk.Button(frm, text='Data Viewer', command=self.open_data_viewer, width=20).pack(side=tk.LEFT)
        # ttk.Button(frm, text='NeXus Browser', command=self.open_file_browser, width=20).pack(side=tk.LEFT)
        ttk.Button(frm, text='Log Viewer', command=self.open_log_viewer, width=20).pack(side=tk.LEFT)
        ttk.Button(frm, text='Notebook Browser', command=self.open_notebook_browser, width=20).pack(side=tk.LEFT)
        ttk.Button(frm, text='Processing', command=self.open_script_runner, width=20).pack(side=tk.LEFT)
        t1 = time()
        logger.info(f"init time: {t1-t0}s")
        # run file-system functions in a thread to speed up the start time
        th = Thread(target=lambda: self.choose_beamline(self.config.get(C.beamline, 'i16')))
        th.start()

    def choose_beamline(self, beamline: str):
        t0 = time()
        bl_config = BEAMLINE_CONFIG[beamline].copy()
        self.config.update(bl_config)
        self.beamline.set('MMG Toolbox: ' + beamline)
        self.visits = get_dls_visits(beamline)
        default_dir = self.config.get(C.default_directory, '.')
        default_dir = default_dir if os.path.isdir(default_dir) else self.config.get(C.recent_data_directories, ['.'])[0]
        self.visits.update({'default': default_dir})
        current_visit = next(iter(self.visits.keys()))
        self.visit_menu.set_menu(current_visit, *self.visits)
        self.visit.set(current_visit)
        t1 = time()
        logger.info(f"choose_beamline time: {t1-t0}s")
        self.dls_directories(self.visits[current_visit])

    def menu_items(self):
        menu = {
            'Recent Files': {
                file: lambda x=file: self.data_dir.set(x)
                for file in self.config.get(C.recent_data_directories)
            },
            'Beamline': {
                bl: lambda x=bl: self.choose_beamline(x)
                for bl in BEAMLINE_CONFIG
            },
        }
        return menu

    def dls_directories(self, data_dir: str):
        t0 = time()
        if not os.access(data_dir, os.R_OK):
            show_error(f"Warning path is not readable: '{data_dir}'", self.root, raise_exception=False)
        self.summary.set(folder_summary_line(data_dir))
        self.data_dir.set(data_dir)
        proc_dir = os.path.join(data_dir, 'processing')
        notebook_dir = os.path.join(data_dir, 'processed', 'notebooks')
        if os.path.isdir(proc_dir) and os.access(proc_dir, os.W_OK):
            self.proc_dir.set(proc_dir)
        if os.path.isdir(notebook_dir) and os.access(notebook_dir, os.R_OK):
            self.notebook_dir.set(notebook_dir)
        t1 = time()
        logger.info(f"dls_directories time: {t1-t0}s")

    def update_config(self):
        save_config(self.config)

    def add_recent_directory(self, directory: str):
        recent = self.config.get(C.recent_data_directories, [])
        if directory in recent:
            return
        recent.insert(0, directory)
        while len(recent) > 10:
            recent.pop()
        self.config[C.recent_data_directories] = recent
        self.update_config()

    def set_current_directories(self):
        self.config[C.current_dir] = self.data_dir.get()
        self.config[C.current_proc] = self.proc_dir.get()
        self.config[C.current_nb] = self.notebook_dir.get()

    def choose_visit(self, event=None):
        visit_folder = self.visits[self.visit.get()]
        self.dls_directories(visit_folder)

    def browse_datadir(self):
        current_folder = self.data_dir.get()
        folder = select_folder(self.root, initial_directory=current_folder if current_folder else None)
        if folder:
            self.dls_directories(folder)

    def browse_analysis(self):
        current_folder = self.proc_dir.get()
        folder = select_folder(self.root, initial_directory=current_folder if current_folder else None)
        if folder:
            self.proc_dir.set(folder)

    def browse_notebook(self):
        current_folder = self.notebook_dir.get()
        folder = select_folder(self.root, initial_directory=current_folder if current_folder else None)
        if folder:
            self.notebook_dir.set(folder)

    def open_data_viewer(self):
        from .. import create_data_viewer
        self.add_recent_directory(self.data_dir.get())
        self.set_current_directories()
        create_data_viewer(self.data_dir.get(), self.root, self.config)

    def open_file_browser(self):
        from .. import create_nexus_file_browser
        create_nexus_file_browser(self.root, self.data_dir.get())

    def open_log_viewer(self):
        from ..apps.log_viewer import create_gda_terminal_log_viewer
        create_gda_terminal_log_viewer(self.data_dir.get(), self.root)

    def open_notebook_browser(self):
        from ..apps.file_browser import create_jupyter_browser
        create_jupyter_browser(self.root, self.notebook_dir.get())

    def open_script_runner(self):
        from ..apps.multi_scan_analysis import create_multi_scan_analysis
        folders = {
            C.default_directory: self.data_dir.get(),
            C.processing_directory: self.proc_dir.get(),
            C.notebook_directory: self.notebook_dir.get(),
        }
        self.config.update(folders)
        create_multi_scan_analysis(self.root, self.config)







"""
mmg_toolbox tests
Test metadata from beamlines
"""

import pytest

from . import only_dls_file_system
from .example_files import DIR

def test_metadata_import():
    errors = None
    try:
        from mmg_toolbox import metadata, xas_metadata, nexus_metadata
    except ImportError as e:
        errors = e
    assert errors is None


@only_dls_file_system
def test_scan_string():
    from mmg_toolbox import Experiment, metadata

    exp = Experiment(DIR + '/i16', instrument='i16')

    # print all scans in folder
    m = f"{metadata.scanno}, {metadata.start}, {metadata.cmd}, {metadata.energy}, {metadata.temp}"
    for scan in exp[:10]:
        scn, start, cmd, energy, temp = scan(m)
        assert scn > 100
        assert len(cmd) > 10

    scan = exp.scan(1109527)
    s = str(scan)
    assert s.count('\n') > 10
    assert 'energy = 7 keV' in s

    energy_str = scan.format('Energy = {incident_energy:.2f} {incident_energy@units?("keV")}')
    assert energy_str == 'Energy = 7.11 keV'
    assert scan.metadata.sy == pytest.approx(-0.7791)
    assert scan.metadata['Atten'] == pytest.approx(30)
    assert scan('Transmission') == pytest.approx(0.005684, abs=1e-6)


@only_dls_file_system
def test_scan_times():
    from mmg_toolbox import data_file_reader
    import datetime

    scan = data_file_reader(DIR + '/i10/i10-921636.nxs')
    start_time, end_time = scan.times('start_time', 'end_time')
    assert isinstance(start_time, datetime.datetime)
    assert isinstance(end_time, datetime.datetime)





"""
a tkinter frame with a single plot
"""
import tkinter as tk

import matplotlib.pyplot as plt
import numpy as np
from numpy import ndarray

from ..misc.config import C
from ..misc.screen_size import get_figure_size
from ..misc.matplotlib import ini_plot
from ..misc.logging import create_logger

logger = create_logger(__file__)


class SimplePlot:
    """
    Simple plot - single plot in frame with axes
    """

    def __init__(self, root: tk.Misc, xdata: list[float], ydata: list[float],
                 xlabel: str = '', ylabel: str = '', title: str = '', config: dict | None = None):
        self.root = root
        self.config = config or {}
        self._y_axis_expansion_factor = 0.1

        fig_size = get_figure_size(root, self.config, C.plot_size)
        self.fig, self.ax1, self.plot_list, self.toolbar = ini_plot(
            frame=self.root,
            figure_size=fig_size,
            figure_dpi=self.config.get(C.plot_dpi, None),
        )
        self.ax1.set_xlabel(xlabel)
        self.ax1.set_ylabel(ylabel)
        self.ax1.set_title(title)
        self.plot(xdata, ydata)

    def plot(self, *args, **kwargs) -> list[plt.Line2D]:
        lines = self.ax1.plot(*args, **kwargs)
        self.plot_list.extend(lines)
        self.update_axes()
        return lines

    def update_labels(self, x_label: str | None = None, y_label: str | None = None,
                      title: str | None = None, legend: bool = False):
        if x_label:
            self.ax1.set_xlabel(x_label)
        if y_label:
            self.ax1.set_ylabel(y_label)
        if title:
            self.ax1.set_title(title)
        if legend:
            self.ax1.legend()
        else:
            self.ax1.legend([]).set_visible(False)

    def plot_from_data(self, x_data: list[ndarray], y_data: list[ndarray], x_label: str = '', y_label: str = '',
                       title: str = '', labels: list[str] | None = None):
        labels = [f"data #{n + 1}" for n in range(len(x_data))] if labels is None else labels
        self.reset_plot()
        for xdata, ydata, label in zip(x_data, y_data, labels):
            lines = self.ax1.plot(np.ravel(xdata), np.ravel(ydata), label=label)
            self.plot_list.extend(lines)
        self.update_labels(x_label=x_label, y_label=y_label, title=title, legend=True if len(labels) > 1 else False)
        self.update_axes()

    def update_from_data(self, x_data: list[ndarray], y_data: list[ndarray], x_label: str | None = None,
                         y_label: str | None = None, title: str | None = None, legend: list[str] | None = None):
        if len(x_data) == len(self.plot_list):
            # replace lines
            legend = [None for _n in range(len(x_data))] if legend is None else legend
            for xdata, ydata, label, line in zip(x_data, y_data, legend, self.plot_list):
                line.set_data(np.ravel(xdata), np.ravel(ydata))
                if label:
                    line.set_label(label)
            self.update_labels(x_label=x_label, y_label=y_label, title=title, legend=True if len(legend) > 1 else False)
            self.update_axes()
        else:
            self.plot_from_data(x_data, y_data, x_label, y_label, title, legend)

    def remove_lines(self):
        for obj in self.plot_list:
            obj.remove()
        self.plot_list.clear()

    def reset_plot(self):
        # self.ax1.set_xlabel(self.xaxis.get())
        # self.ax1.set_ylabel(self.yaxis.get())
        # self.ax1.set_title('')
        self.ax1.set_prop_cycle(None)  # reset colours
        self.ax1.legend([]).set_visible(False)
        self.remove_lines()

    def _relim(self):
        if not any(len(line.get_xdata()) for line in self.plot_list):
            return
        max_x_val = max(np.max(x) for line in self.plot_list if len(x := line.get_xdata()) > 0)
        min_x_val = min(np.min(x) for line in self.plot_list if len(x := line.get_xdata()) > 0)
        max_y_val = max(np.max(y) for line in self.plot_list if len(y := line.get_ydata()) > 0)
        min_y_val = min(np.min(y) for line in self.plot_list if len(y := line.get_ydata()) > 0)
        # expand y-axis slightly beyond data
        y_diff = max_y_val - min_y_val
        if y_diff == 0:
            y_diff = max_y_val + 0.01
        y_axis_max = max_y_val + self._y_axis_expansion_factor * y_diff
        y_axis_min = min_y_val - self._y_axis_expansion_factor * y_diff
        # max_y_val = 1.05 * max_y_val if max_y_val > 0 else max_y_val * 0.98
        # min_y_val = 0.95 * min_y_val if min_y_val > 0 else min_y_val * 1.02
        self.ax1.axis((min_x_val, max_x_val, y_axis_min, y_axis_max))
        self.ax1.autoscale_view()

    def update_axes(self):
        # self.ax1.relim()
        # self.ax1.autoscale(True)
        # self.ax1.autoscale_view()
        self._relim()
        self.fig.canvas.draw()
        self.toolbar.update()


class MultiAxisPlot(SimplePlot):
    def __init__(self, root: tk.Misc, xdata: list[float], ydata: dict[str, float],
                 xlabel: str = '', ylabel: str = '', title: str = '', config: dict | None = None):
        #TODO: Complete multi-axis plot, use ideas from nexus_plot.py
        super().__init__(root, xdata, ydata, xlabel, ylabel, title, config)



"""
Test autoprocessing notebooks using papermill
"""

import os
import h5py
from . import only_dls_file_system
from .example_files import FILES_DICT

NOTEBOOKS = os.path.join(os.path.dirname(__file__), '..', 'notebooks')
NB_PATHS = {
    name: os.path.join(NOTEBOOKS, name)
    for name in os.listdir(NOTEBOOKS) if name.endswith('.ipynb')
}


@only_dls_file_system
def test_nexus_processor():
    import papermill as pm
    import nbformat

    dat_file = '/dls/science/groups/das/ExampleData/hdfmap_tests/i10/spool/i10-618365.dat'
    if os.path.isfile(dat_file):
        os.remove(dat_file)

    pm.execute_notebook(
        NB_PATHS['nexus_processor.ipynb'],
        'output.ipynb',
        parameters={
            'inpath': FILES_DICT['i10 scan'],
            'outpath': 'output.nxs',
        }
    )

    nb = nbformat.read('output.ipynb', as_version=4)
    assert nb.metadata.papermill.exception is None
    assert os.path.isfile(dat_file)

    os.remove('output.ipynb')
    os.remove(dat_file)


@only_dls_file_system
def test_nexus2srs_processor():
    import papermill as pm
    import nbformat

    dat_file = '/dls/science/groups/das/ExampleData/hdfmap_tests/i10/spool/i10-921636.dat'
    if os.path.isfile(dat_file):
        os.remove(dat_file)

    pm.execute_notebook(
        NB_PATHS['nexus2srs_processor.ipynb'],
        'output.ipynb',
        parameters={
            'inpath': FILES_DICT['i10 pimte new scan'],
            'outpath': 'output.nxs',
        }
    )

    nb = nbformat.read('output.ipynb', as_version=4)
    assert nb.metadata.papermill.exception is None
    assert os.path.isfile(dat_file)

    os.remove('output.ipynb')
    os.remove(dat_file)


@only_dls_file_system
def test_autoprocess_xas_notebook():
    import papermill as pm
    pm.execute_notebook(
        NB_PATHS['xas_notebook.ipynb'],
        'output.ipynb',
        parameters={
            'inpath': FILES_DICT['i06-1 zacscan'],
            'outpath': 'output.nxs',
        }
    )
    assert os.path.isfile('output.ipynb')
    assert os.path.isfile('output.nxs')

    with h5py.File('output.nxs', 'r') as hdf:
        assert isinstance(hdf['/entry/divide_by_preedge/tey'], h5py.Dataset)

    os.remove('output.ipynb')
    os.remove('output.nxs')


@only_dls_file_system
def test_autoprocess_xmcd_processor():
    print("No example xmcd_processor file, skipping...")
    assert os.path.isfile(NB_PATHS['xmcd_processor.ipynb'])
    # import papermill as pm
    # pm.execute_notebook(
    #     NB_PATHS['xmcd_processor.ipynb'],
    #     'output.ipynb',
    #     parameters={
    #         'inpath': FILES_DICT[''],  # need example with scan x -10 -1 1 xmcd_processor
    #         'outpath': 'output.nxs',
    #     }
    # )
    # assert os.path.isfile('output.ipynb')
    # assert os.path.isfile('1234-1235_xmcd.nxs')
    #
    # with h5py.File('output.nxs', 'r') as hdf:
    #     assert isinstance(hdf['/processed/xmcd/tey'], h5py.Dataset)
    #
    # os.remove('output.ipynb')
    # os.remove('output.nxs')


@only_dls_file_system
def test_msmapper_processor():
    import papermill as pm
    import nbformat

    # Remove analysis group from output if it exists
    with h5py.File(FILES_DICT['msmapper volume 527'], 'a') as hdf:
        if 'analysis' in hdf:
            del hdf['analysis']

    pm.execute_notebook(
        NB_PATHS['msmapper_processor.ipynb'],
        'output.ipynb',
        parameters={
            'inpath': FILES_DICT['i16 pilatus eta scan, new nexus format'],
            'outpath': 'output.nxs',
        }
    )

    nb = nbformat.read('output.ipynb', as_version=4)
    assert nb.metadata.papermill.exception is None

    with h5py.File(FILES_DICT['msmapper volume 527']) as hdf:
        assert isinstance(hdf['/analysis/h_axis/fit'], h5py.Dataset)

    os.remove('output.ipynb')



"""
mmg_toolbox tests
Test top-level imports
"""

def test_standard_imports():
    errors = None
    try:
        import mmg_toolbox
        from mmg_toolbox import data_file_reader, Experiment, version_info, module_info
    except ImportError as e:
        errors = e
    assert errors is None


def test_xas_imports():
    errors = None
    try:
        from mmg_toolbox import xas
        from mmg_toolbox.utils.file_functions import replace_scan_number
        from mmg_toolbox.xas.nxxas_loader import is_nxxas
        from mmg_toolbox.xas.spectra_container import average_polarised_scans
        from mmg_toolbox.utils.gda_functions import gda_datavis_file_message
    except ImportError as e:
        errors = e
    assert errors is None


def test_fitting_imports():
    errors = None
    try:
        from mmg_toolbox import fitting
        from mmg_toolbox.fitting import multipeakfit, peakfit
        # from mmg_toolbox.utils import fitting
    except ImportError as e:
        errors = e
    assert errors is None


def test_plotting_imports():
    errors = None
    try:
        from mmg_toolbox import plotting
        from mmg_toolbox.plotting.matplotlib import set_plot_defaults
    except ImportError as e:
        errors = e
    assert errors is None


def test_other_imports():
    errors = None
    try:
        from mmg_toolbox.diffraction.lattice import bmatrix, cal2theta
        from mmg_toolbox.diffraction.msmapper import update_msmapper_nexus
    except ImportError as e:
        errors = e
    assert errors is None




"""
tkinter styles

Uses ttk Themed Styles

Docs:
https://tkdocs.com/tutorial/styles.html
List of Themes:
https://wiki.tcl-lang.org/page/List+of+ttk+Themes
awlight/awdark theme download:
https://wiki.tcl-lang.org/page/awthemes
Fonts:
https://tkdocs.com/tutorial/fonts.html
"""
from __future__ import annotations

import os
import tkinter as tk
from tkinter import ttk
from .logging import create_logger

logger = create_logger(__name__)

AWTHEMES_PATH = '/dls/science/users/grp66007/tkinter/awthemes-10.4.0'
DEFAULT_THEME = 'clam'

def create_style(root):
    style = ttk.Style(root)
    if os.path.isdir(AWTHEMES_PATH):
        root.tk.call('lappend', 'auto_path', AWTHEMES_PATH)
        root.tk.call('package', 'require', 'awdark')
        root.tk.call('package', 'require', 'awlight')
    use_theme(style, DEFAULT_THEME)
    return style


def extra_styles(style: ttk.Style):
    """
    Add additional styles to use on individual ttk components
    """
    # style.configure("Red.TLabel", foreground='red', font='TkDefaultFont 24 bold')
    style.configure("title.TLabel", foreground='red', font='times 24 bold')
    style.configure("subtitle.TLabel", foreground='black', font='times 18 bold')
    style.configure("error.TLabel", foreground='red')
    style.configure("smallMsg.TLabel", font='TkDefaultFont 10 bold')


def use_theme(style: ttk.Style, theme_name: str):
    if theme_name in style.theme_names():
        style.theme_use(theme_name)
    else:
        logger.warning(f"ttk Style Theme not available: '{theme_name}'")
    extra_styles(style)


def theme_menu(style: ttk.Style) -> dict:
    """Generate theme menu dict"""
    menu = {
        "Themes": {
            name: lambda n=name: use_theme(style, n)
            for name in style.theme_names()
        }
    }
    return menu


class RootWithStyle(tk.Tk):
    style: ttk.Style | None = None


def create_root(window_title: str, parent: tk.Misc | RootWithStyle | None = None) -> RootWithStyle:
    """Create tkinter root object with style attribute"""
    # TODO: this is the slow part of the startup, investigate further.
    if parent:
        root = tk.Toplevel(parent)
        # root.geometry(f"+{parent.winfo_x()+100}+{parent.winfo_y()+100}")
        # root.transient(parent)
        if hasattr(parent, 'style'):
            root.style = parent.style
    else:
        root = tk.Tk()

    def update(event):
        root.update()
        print(root.winfo_reqwidth(), root.winfo_screenheight())
        print(root.winfo_screenwidth(), root.winfo_screenheight())

    # root.bind('<Configure>', update)

    if not hasattr(root, 'style'):
        style = create_style(root)
        root.style = style

    # Fix background (on windows)
    bg = get_style_background(root, root.style)
    root.configure(bg=bg)

    root.wm_title(window_title)
    # self.root.minsize(width=640, height=480)
    # root.maxsize(width=root.winfo_screenwidth() * 3 // 4, height=root.winfo_screenheight() * 3 // 4)
    root.maxsize(width=int(root.winfo_screenwidth() - 50), height=int(root.winfo_screenheight()) - 100)
    return root


def create_hover(parent: tk.Misc | RootWithStyle, top_left: tuple[float, float] = (0.1, 0.1)):
    """
    Create tkinter frame hovering above the current widget

    E.G.
    window_frame, close = create_hover(widget)
    ttk.Button(window_frame, text='Close', command=close).pack()

    :param parent: tk widget or root
    :param top_left: (relx, rely) widget top-left corner relative to parent top-left corner
    :returns: ttk.Frame object inside tk.TopLevel with no window management
    :returns: function close() -> None (releases widget and destroys hover window)
    """

    root = ttk.Frame(parent)
    root.place(relx=top_left[0], rely=top_left[1])

    window = ttk.Frame(root, borderwidth=20, relief=tk.RAISED)
    window.pack(side=tk.TOP, fill=tk.BOTH)

    def destroy(event=None):
        root.grab_release()
        root.destroy()

    root.grab_set()
    return window, destroy


def update_text_style(widget: tk.Text, style: ttk.Style):
    """
    Update a tk.Text widget with the current style
    """
    widget.configure(
        bg=style.lookup('.', 'background'),
        fg=style.lookup('.', 'foreground'),
        font=style.lookup('.', 'font') or 'TkDefaultFont',
    )


def get_style_background(widget: tk.Misc, style: ttk.Style | None = None) -> str:
    """
    Return the background colour of the current style
    """
    style = style or ttk.Style()

    # Get the system color name
    color_name = style.lookup('.', 'background')  # e.g., 'SystemButtonFace'

    # Convert to RGB (16-bit per channel)
    r16, g16, b16 = widget.winfo_rgb(color_name)

    # Normalize to 8-bit and format as hex
    r, g, b = (r16 // 256, g16 // 256, b16 // 256)
    hex_color = f"#{r:02x}{g:02x}{b:02x}"
    return hex_color


# from: https://stackoverflow.com/questions/45389166/how-to-know-all-style-options-of-a-ttk-widget

def _iter_layout(layout, tab_amnt=0, elements=[]):
    """Recursively prints the layout children."""
    el_tabs = '  '*tab_amnt
    val_tabs = '  '*(tab_amnt + 1)

    for element, child in layout:
        elements.append(element)
        print(el_tabs+ '\'{}\': {}'.format(element, '{'))
        for key, value in child.items():
            if type(value) == str:
                print(val_tabs + '\'{}\' : \'{}\','.format(key, value))
            else:
                print(val_tabs + '\'{}\' : [('.format(key))
                _iter_layout(value, tab_amnt=tab_amnt+3)
                print(val_tabs + ')]')

        print(el_tabs + '{}{}'.format('} // ', element))

    return elements


def stylename_elements_options(widget: tk.Misc):
    """
    Function to expose the options of every element associated to a widget stylename.

        widget = ttk.Button(None)
        class_ = widget.winfo_class()
        stylename_elements_options(class_, widget)
    """

    try:
        # Get widget elements
        stylename = widget.winfo_class()
        style = ttk.Style()
        layout = style.layout(stylename)
        config = widget.configure()

        print('{:*^50}\n'.format(f'Style = {stylename}'))

        print('{:*^50}'.format('Config'))
        for key, value in config.items():
            print('{:<15}{:^10}{}'.format(key, '=>', value))

        print('\n{:*^50}'.format('Layout'))
        elements = _iter_layout(layout)

        # Get options of widget elements
        print('\n{:*^50}'.format('element options'))
        for element in elements:
            print('{0:30} options: {1}'.format(
                element, style.element_options(element)))

    except tk.TclError:
        print('_tkinter.TclError: "{0}" in function'
                'widget_elements_options({0}) is not a regonised stylename.'
                .format(stylename))




"""
Environment functions
"""

import sys
import os
import re
import subprocess
import tempfile
from datetime import datetime

from mmg_toolbox.utils.file_functions import get_scan_number, list_files

# environment variables on beamline computers
BEAMLINE = 'BEAMLINE'
USER = ['USER', 'USERNAME']
DLS = '/dls'

# Find writable directory
TMPDIR = tempfile.gettempdir()
if not os.access(TMPDIR, os.W_OK):
    TMPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    if not os.access(TMPDIR, os.W_OK):
        TMPDIR = os.path.expanduser('~')


# Initialise available beamlines
YEAR = str(datetime.now().year)


def check_file_access(filepath: str, append: str = '_new') -> str:
    """Check path has write access, if not, return appended path with write access"""
    # return filepath if it already exists and is writable
    if os.path.exists(filepath) and os.access(filepath, os.W_OK):
        return filepath
    tries = 0
    max_tries = 3

    path, name = os.path.split(filepath)
    if not os.access(path, os.W_OK):
        raise OSError(f"new file cannot be written as path is not writable: '{path}'")

    if os.path.exists(filepath):
        # filepath is not writeable, amend name
        name, ext = os.path.splitext(name)
        while os.path.exists(filepath) and not os.access(filepath, os.W_OK):
            if tries > max_tries:
                raise Exception(f"File is not writable: {filepath}")
            name += append
            filepath = os.path.join(path, name + ext)
            tries += 1
    return filepath


def get_beamline(default='', filename: str | None = None) -> str:
    """Return current beamline from filename/filepath or environment variable"""
    env_bl = os.environ.get(BEAMLINE, default)
    if filename is None:
        return env_bl
    return get_beamline_from_directory(filename, env_bl)


def get_beamline_from_directory(directory: str, default: str = ''):
    """Return current beamline from given directory"""
    beamlines = re.findall('/([a-zA-Z][0-9]{2}-?[1-9]?)', directory)
    return beamlines[0] if beamlines else default


def get_user(default=''):
    """Return current user from environment variable"""
    return next((os.environ[u] for u in USER if u in os.environ), default)


def get_data_directory():
    """Return the default data directory"""
    beamline = get_beamline()
    year = datetime.now().year
    if beamline:
        return f"/dls/{beamline}/data/{year}"
    return os.path.expanduser('~')


def get_processing_directory(data_directory: str):
    """Return the processing directory of the visit"""
    return os.path.join(data_directory, 'processing')


def get_notebook_directory(data_directory: str):
    """Return the notebook directory of the visit"""
    return os.path.join(data_directory, 'processed', 'notebooks')


def get_dls_visits(instrument: str | None = None, year: str | int | None = None) -> dict[str, str]:
    """Return dict of {visit: path} for each visible visit in the beamline directory"""
    if instrument is None:
        instrument = get_beamline()
    if year is None:
        year = datetime.now().year

    dls_dir = os.path.join(DLS, instrument.lower(), 'data', str(year))
    if os.path.isdir(dls_dir):
        return {
            os.path.basename(path): path
            for path in sorted(
                (file.path for file in os.scandir(os.path.join(DLS, instrument, 'data', YEAR))
                 if file.is_dir() and os.access(file.path, os.R_OK)),
                key=lambda x: os.path.getmtime(x), reverse=True
            )
        }
    return {}


def get_first_file(folder: str, extension='.nxs') -> str:
    """Return first scan in folder"""
    return next(iter(list_files(folder, extension=extension)))


def get_scan_numbers(folder: str, extension: str = '.nxs') -> list[int]:
    """Return ordered list of scans numbers from nexus files in directory"""
    return sorted(
        number for filename in list_files(folder, extension=extension)
        if (number := get_scan_number(filename)) > 0
    )


def scan_number_mapping(*folders: str, extension: str = '.nxs') -> dict[int, str]:
    """Build mapping of scan number to scan file"""
    mapping = {
        number: filename
        for folder in folders
        for filename in list_files(folder, extension=extension)
        if (number := get_scan_number(filename)) > 0
    }
    return dict(sorted(mapping.items()))


def get_last_scan_number(folder: str) -> int:
    """Return latest scan number"""
    return get_scan_numbers(folder)[-1]


def last_folder_update(folder: str) -> datetime:
    """Returns datetime timestamp of last folder update"""
    modified = os.path.getmtime(folder)
    return datetime.fromtimestamp(modified)


def get_scan_notebooks(scan: int | str, data_directory: str | None = None) -> list[str]:
    """Return list of processed jupyter notebooks for scan"""
    try:
        data_directory, filename = os.path.split(scan)
        scan = get_scan_number(filename)
    except TypeError:
        pass
    notebook_directory = get_notebook_directory(data_directory)
    if os.path.isdir(notebook_directory):
        notebooks = list_files(notebook_directory, '.ipynb')
        return [notebook for notebook in notebooks if str(scan) in os.path.basename(notebook)]
    return []


def run_command(command: str):
    """
    Run shell command, print output to terminal
    """
    print('\n\n\n################# Starting ###################')
    print(f"Running command:\n{command}\n\n\n")
    output = subprocess.run(command, shell=True, capture_output=True)
    print(output.stdout.decode())
    print(output.stderr.decode())
    print('\n\n\n################# Finished ###################\n\n\n')


def open_terminal(command: str):
    """
    Open a new terminal window (linux only) and run a command
    """
    shell_cmd = f"gnome-terminal -- bash -c \"{command}; exec bash\""
    subprocess.Popen(shell_cmd, shell=True)


def run_python_script(script_filename: str):
    """
    Run shell command, print output to terminal
    """
    command = f"{sys.executable} {script_filename}"
    run_command(command)


def run_python_string(script: str):
    """
    Run shell command, print output to terminal
    """
    # command = f"{sys.executable} -c {script}"
    # run_command(command)
    subprocess.run([sys.executable, '-c', script])


def run_jupyter_notebook(notebook_filename: str):
    """
    Run a jupyter notebook
    """
    command = f"jupyter notebook {notebook_filename}"
    run_command(command)


def open_jupyter_lab():
    """
    Open a new terminal and start a Jupyter lab terminal (linux only)
    """
    shell_cmd = f"gnome-terminal -- bash -c \"jupyter lab; exec bash\""
    subprocess.Popen(shell_cmd, shell=True)


"""
A treeview tkinter frame for displaying the hierachical structure of HDF and Nexus files
"""

import os
import h5py
import tkinter as tk

import hdfmap
from hdfmap.eval_functions import generate_identifier

from .treeview import CanvasTreeview
from ..misc.logging import create_logger

logger = create_logger(__file__)


class HdfTreeview(CanvasTreeview):
    """
    HDF Treeview object
    """
    def __init__(self, root: tk.Misc, width: int | None = None, height: int | None = None):
        columns = [
            ('#0', 'HDF Address', 400, False, None),
            ('type', 'Type', 100, False, None),
            ('name', 'Name', 100, False, None),
            ('value', 'Value', 200, False, None),
        ]
        super().__init__(root, *columns, width=width, height=height)


    def populate(self, hdf_obj: h5py.File, openstate=True):
        """Load HDF file, populate ttk.treeview object"""

        def recur_func(hdf_group, tree_group="", top_address='/'):
            for key in hdf_group:
                obj = hdf_group.get(key)
                link = hdf_group.get(key, getlink=True)
                address = top_address + key
                name = generate_identifier(address)
                if isinstance(obj, h5py.Group):
                    try:
                        nx_class = obj.attrs['NX_class'].decode() if 'NX_class' in obj.attrs else 'Group'
                    except AttributeError:
                        nx_class = obj.attrs['NX_class']
                    except OSError:
                        nx_class = 'Group'  # if object doesn't have attrs
                    values = (nx_class, name, "")
                    new_tree_group = self.tree.insert(tree_group, tk.END, text=address, values=values)
                    # add attributes
                    for attr, val in obj.attrs.items():
                        self.tree.insert(new_tree_group, tk.END, text=f"@{attr}", values=('Attribute', attr, val))
                    recur_func(obj, new_tree_group, address + '/')
                    self.tree.item(new_tree_group, open=openstate)
                elif isinstance(obj, h5py.Dataset):
                    if isinstance(link, h5py.ExternalLink):
                        link_type = 'External Link'
                    elif isinstance(link, h5py.SoftLink):
                        link_type = 'Soft Link'
                    else:
                        link_type = 'Dataset'
                    if obj.shape:
                        val = f"{obj.dtype} {obj.shape}"
                    else:
                        val = str(obj[()])
                    values = (link_type, name, val)
                    # datasets.append(address)
                    new_tree_group = self.tree.insert(tree_group, tk.END, text=address, values=values)
                    for attr, val in obj.attrs.items():
                        self.tree.insert(new_tree_group, tk.END, text=f"@{attr}", values=('Attribute', attr, val))
                    self.tree.item(new_tree_group, open=False)

        # add top level file group
        hdf_filename = hdf_obj.filename
        self.tree.insert("", tk.END, text='/', values=('File', os.path.basename(hdf_filename), ''))
        recur_func(hdf_obj, "")


class HdfNameSpace(CanvasTreeview):
    """
    HDF Namespace object
    """

    def __init__(self, root: tk.Misc, width: int | None = None, height: int | None = None):
        columns = [
            ('#0', 'Name', 100, False, None),
            ('path', 'Path', 300, False, None),
            ('value', 'Value', 200, False, None),
        ]
        super().__init__(root, *columns, width=width, height=height)

    def populate(self, hdf_obj: h5py.File, hdf_map: hdfmap.NexusMap,
                 all: bool = True, group: bool = False, combined: bool = False, values: bool = False,
                 arrays: bool = False, scannables: bool = False, metadata: bool = False, image_data: bool = False):
        """Load HDF file, populate ttk.treeview object"""

        data = {
            name: hdf_map.get_string(hdf_obj, name)
            for name, path in hdf_map.combined.items()
        }

        if all or group:
            datasets = self.tree.insert("", tk.END, text='Groups', values=('', ''))
            for name, path_list in hdf_map.classes.items():
                # path_list = list(set(path_list))  # remove duplicates
                if len(path_list) == 1:
                    self.tree.insert(datasets, tk.END, text=name, values=(path_list[0], ''))
                else:
                    grp = self.tree.insert(datasets, tk.END, text=name, values=('', ''))
                    for path in path_list:
                        self.tree.insert(grp, tk.END, text=name, values=(path, ''))

        if all or combined:
            datasets = self.tree.insert("", tk.END, text='Combined', values=('', ''))
            for name, path in hdf_map.combined.items():
                value = data.get(name, 'NOT IN MAP')
                self.tree.insert(datasets, tk.END, text=name, values=(path, value))

        if all or values:
            datasets = self.tree.insert("", tk.END, text='Values', values=('', ''))
            for name, path in hdf_map.values.items():
                value = data.get(name, 'NOT IN MAP')
                self.tree.insert(datasets, tk.END, text=name, values=(path, value))

        if all or arrays:
            datasets = self.tree.insert("", tk.END, text='Arrays', values=('', ''))
            for name, path in hdf_map.arrays.items():
                value = data.get(name, 'NOT IN MAP')
                self.tree.insert(datasets, tk.END, text=name, values=(path, value))

        if all or scannables:
            datasets = self.tree.insert("", tk.END, text='Scannables', values=('', ''))
            for name, path in hdf_map.scannables.items():
                value = data.get(name, 'NOT IN MAP')
                self.tree.insert(datasets, tk.END, text=name, values=(path, value))

        if all or metadata:
            datasets = self.tree.insert("", tk.END, text='Metadata', values=('', ''))
            for name, path in hdf_map.metadata.items():
                value = data.get(name, 'NOT IN MAP')
                self.tree.insert(datasets, tk.END, text=name, values=(path, value))

        if all or image_data:
            datasets = self.tree.insert("", tk.END, text='Image Data', values=('', ''))
            for name, path in hdf_map.image_data.items():
                value = data.get(name, 'NOT IN MAP')
                self.tree.insert(datasets, tk.END, text=name, values=(path, value))




"""
Test nx_find
"""

import h5py

from . import only_dls_file_system
from .example_files import DIR
from mmg_toolbox.nexus.nexus_functions import nx_find, nx_find_all


@only_dls_file_system
def test_nx_find():
    f = DIR + 'i06/i06-1-372210.nxs'

    with h5py.File(f) as hdf:

        dataset = nx_find(hdf, 'NXdata', 'signal')
        assert isinstance(dataset, h5py.Dataset)
        assert dataset.shape == (500, )

        dataset = nx_find(hdf, 'axes')
        assert isinstance(dataset, h5py.Dataset)
        assert dataset.shape == (500, )
        assert dataset.name == '/entry/fesData/fastEnergy'

        datasets = nx_find_all(hdf, 'axes')
        assert isinstance(datasets, list)
        assert len(datasets) == 6

        dataset = nx_find(hdf, 'NXentry', 'instrument/s1/x_gap')
        assert isinstance(dataset, h5py.Dataset)
        assert dataset.name == '/entry/instrument/s1/x_gap'



"""
mmg_toolbox tests
Test lmfit functions
"""
import pytest
import numpy as np
from lmfit.models import GaussianModel

from mmg_toolbox import data_file_reader
from mmg_toolbox.fitting import FitResults, peakfit, multipeakfit, gauss, Peak

from . import only_dls_file_system
from .example_files import DIR

@pytest.fixture
def example_peak():
    x = np.linspace(-3, 2, 100)
    y = gauss(x, height=10, cen=-0.5, fwhm=0.8, bkg=0.1)
    yield x, y


def test_gauss():
    x = np.linspace(-1, 1, 100)
    h, c, w, b = 10, 0, 0.2, 0
    amplitude = h * w * 1.064467  # area
    sigma = w / 2.348200
    g = gauss(x, height=h, cen=c, fwhm=w, bkg=b)

    g_area = np.trapezoid(g, x)
    g_height = np.max(g)
    assert pytest.approx(g_area, abs=1e-2) == amplitude
    assert pytest.approx(g_height, abs=0.1) == h

    g_model = GaussianModel()
    pars = g_model.make_params(amplitude=amplitude, center=c, sigma=sigma)
    g_lmfit = g_model.eval(pars, x=x)

    g_lm_area = np.trapezoid(g_lmfit, x)
    g_lm_height = np.max(g_lmfit)
    assert pytest.approx(g_lm_area, abs=1e-2) == amplitude
    assert pytest.approx(g_lm_height, abs=0.1) == h

    # Check integration
    area = np.sum(g) * np.mean(np.diff(x))
    assert pytest.approx(area, abs=1e-2) == amplitude

    # Check Gauss2D
    y = np.linspace(-1, 1, 100)
    g2d = gauss(x, y, height=h, cen=c, fwhm=w, bkg=b)
    amplitude2 = h * w * w * (np.pi / (4*np.log(2)))
    integral2 = np.trapezoid(np.trapezoid(g2d, y, axis=0), x, axis=0)
    assert g2d.shape == (100, 100)
    area_2d = np.sum(g2d) * np.mean(np.diff(x))  * np.mean(np.diff(y))
    height_2d = np.max(g2d)
    assert pytest.approx(area_2d, abs=1e-2) == amplitude2
    assert pytest.approx(area_2d, abs=1e-2) == integral2
    assert pytest.approx(height_2d, abs=0.15) == h

    # Check Gauss2D with different values
    y = np.linspace(-10, -2, 300)
    h, c, c2, w, w2, b = 10, -0.1, -5, 0.2, 0.5, 0
    g2d = gauss(x, y, height=h, cen=c, fwhm=w, bkg=b, cen_y=c2, fwhm_y=w2)
    amplitude2 = h * w * w2 * (np.pi / (4 * np.log(2)))
    integral2 = np.trapezoid(np.trapezoid(g2d, y, axis=0), x, axis=0)
    assert g2d.shape == (300, 100)
    area_2d = np.sum(g2d) * np.mean(np.diff(x)) * np.mean(np.diff(y))
    height_2d = np.max(g2d)
    assert pytest.approx(area_2d, abs=1e-2) == amplitude2
    assert pytest.approx(area_2d, abs=1e-2) == integral2
    assert pytest.approx(height_2d, abs=0.15) == h





def test_peak_fit(example_peak):
    x, y = example_peak
    result = peakfit(x, y)
    print(result)
    assert isinstance(result, FitResults)
    assert abs(result.height - 10) < 1
    assert abs(result.center + 0.5) < 0.01
    assert abs(result.fwhm - 0.8) < 0.1
    assert abs(result.background - 0.1) < 1
    assert result.amplitude > 3 * result.stderr_amplitude


def test_multipeakfit(example_peak):
    x, y = example_peak
    result = multipeakfit(x, y)
    assert isinstance(result, FitResults)
    assert result.npeaks == 1
    assert abs(result.height - 10) < 1
    assert abs(result.center + 0.5) < 0.01
    assert abs(result.fwhm - 0.8) < 0.1
    assert abs(result.background - 0.1) < 1
    assert result.amplitude > 3 * result.stderr_amplitude

    for peak in result:
        print(peak)
        assert isinstance(peak, Peak)
        assert result.amplitude == pytest.approx(8.52, abs=0.01)


@only_dls_file_system
def test_scan_fit():
    file = DIR + f'/i16/777777.nxs'
    scan = data_file_reader(file)

    result = scan.fit.multi_peak_fit('eta', 'sum', model='pVoight')
    assert result.npeaks == len(result) == 5
    assert result.amplitude == pytest.approx(1.3493e6, abs=1e3)



"""
Simple frame widget that returns the current screen size and window position
"""

from ..misc.styles import tk, ttk, create_root
from ..misc.logging import create_logger

logger = create_logger(__file__)


class WindowSize:
    """
    Simple text edit box
        next_text = EditText(old_text, tkframe).show()
    :param parent: tk root
    """

    def __init__(self, parent: tk.Misc | None = None):
        self.root = create_root('Size Widget', parent=parent)

        frm = ttk.LabelFrame(self.root, text='Size')
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        self.text = tk.Text(frm, wrap=tk.NONE, width=100, height=10)
        self.text.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)

        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        var = ttk.Button(self.root, text='Update', command=self.fun_update)
        var.pack(side=tk.TOP, fill=tk.X)
        var = ttk.Button(self.root, text='Close', command=self.root.destroy)
        var.pack(side=tk.TOP, fill=tk.X)

    def fun_update(self, event=None):
        """Launches window, returns selection"""
        self.root.update()
        self.root.update_idletasks()
        # self.root.attributes('-fullscreen', True)
        # self.root.state('iconic')
        # geometry = self.root.winfo_geometry()
        s = f"Screen size (w x h): {self.root.winfo_screenwidth()} x {self.root.winfo_screenheight()}\n"
        s += f"Geometry: {self.root.geometry()}\n"
        s += f"winfo_reqwidth x winfo_reqhight: {self.root.winfo_reqwidth()} x {self.root.winfo_reqheight()}\n"
        # s += f"current display geometry: {self.get_curr_screen_geometry()}"
        self.text.replace("1.0", tk.END, s)




"""
mmg_toolbox tests
Test data readers
"""

import time
from mmg_toolbox.nexus.nexus_scan import NexusScan
from . import only_dls_file_system
from .example_files import DIR, FILES


@only_dls_file_system
def test_multi_expression_time():
    # Check speed of reading a file multiple times vs opening once
    f, desc = FILES[4]
    scan = NexusScan(f)
    names = list(scan.map.combined.keys())[:100]
    t0 = time.time_ns()
    values_multi = [scan.eval(name) for name in names]
    t1 = time.time_ns()
    with scan.map.load_hdf() as hdf:
        values_single = [scan.map.eval(hdf, name) for name in names]
    t2 = time.time_ns()
    values_multi = [scan.eval(name) for name in names]
    t3 = time.time_ns()

    print(f"\nMulti-read time = {(t1 - t0) * 1e-9:.3f} seconds")
    print(f"Single-read time = {(t2 - t1) * 1e-9:.3f} seconds")
    print(f"Multi-read time = {(t3 - t2) * 1e-9:.3f} seconds")
    ratio = (t3 - t2) / (t2 - t1)
    print(f"Single read time is {ratio:.3g} times faster then multi-read")
    assert values_multi == values_single





"""
tk widgets shows recent vists for beamlines
"""

import os
import tkinter as tk
from tkinter import ttk
from datetime import datetime, timezone
from threading import Thread


from mmg_toolbox.utils.env_functions import get_notebook_directory, get_dls_visits, scan_number_mapping
from ..apps.data_viewer import create_data_viewer
from ..misc.logging import create_logger
from ..misc.config import get_config

logger = create_logger(__file__)


def visit_frame(parent: tk.Misc, visit: str, scans: dict[int, str], notebooks: dict[int, str], command: () = None):

    ttk.Button(parent, text=visit, width=12, command=command).pack(side="left", padx=5)

    frm = ttk.Frame(parent)
    frm.pack(side="left", fill="y", expand=True)

    if len(scans) > 0:
        last_scan_number = list(scans)[-1]
        last_scan_path = list(scans.values())[-1]
        last_scan_time = datetime.fromtimestamp(os.path.getmtime(last_scan_path), tz=timezone.utc)
        last_scan = f", #{last_scan_number:<8} ({last_scan_time:%Y-%m-%d %H:%M})"
    else:
        last_scan = ""
    scan_str = f"Scans: {len(scans):4}{last_scan}"
    ttk.Label(frm, text=scan_str).pack(side="top", fill="x", padx=5)

    if len(notebooks) > 0:
        last_notebook_number = list(notebooks)[-1]
        last_notebook_path = list(notebooks.values())[-1]
        last_notebook_time = datetime.fromtimestamp(os.path.getmtime(last_notebook_path), tz=timezone.utc)
        last_notebook = f", #{last_notebook_number:<8} ({last_notebook_time:%Y-%m-%d %H:%M})"
    else:
        last_notebook = ""
    nb_str = f"Notebooks: {len(notebooks):4}{last_notebook}"
    ttk.Label(frm, text=nb_str).pack(side="top", fill="x", padx=5)


class InstrumentVisits:
    def __init__(self, parent: tk.Misc, instrument: str, config: dict | None = None):
        self.root = parent
        self.config = config or get_config()
        self.instrument = instrument
        self.visits = get_dls_visits(instrument)
        self.last_visits = list(self.visits)[:3]

        self.frame = ttk.Frame(parent, relief="raised", borderwidth=2)
        self.frame.pack(side="top", fill="both", expand=True)
        ttk.Label(self.frame, text=instrument, style="title.TLabel").grid(row=0, column=0, columnspan=3, padx=5, pady=20, sticky="w")

        th = Thread(target=self.populate_visits)
        th.start()

    def populate_visits(self):
        r, c = 0, 0
        for visit, path in self.visits.items():
            cmd = lambda p=path: create_data_viewer(path, self.root, self.config)

            if r < 3:
                scans = scan_number_mapping(path)
                if len(scans) > 0:
                    notebooks = scan_number_mapping(get_notebook_directory(path), extension='.ipynb')
                    line = ttk.Frame(self.frame)
                    line.grid(row=r+1, column=0, columnspan=3, padx=5, pady=5, sticky="ew")
                    visit_frame(line, visit, scans, notebooks, cmd)
                    r += 1
            elif r < 5:
                ttk.Button(self.frame, text=visit, width=12, command=cmd).grid(row=r+1, column=c, padx=5, pady=5)
                r += 1 if c == 2 else 0
                c = c + 1 if c < 2 else 0






from .hdfmap_generic import HdfMapMMGMetadata, HdfMapXASMetadata, HdfMapNexus

__all__ = ['metadata', 'xas_metadata', 'nexus_metadata']

metadata = HdfMapMMGMetadata
xas_metadata = HdfMapXASMetadata
nexus_metadata = HdfMapNexus





from tkinter import ttk

from mmg_toolbox.beamline_metadata.config import BEAMLINE_CONFIG
from ..misc.config import get_config
from ..misc.styles import create_root


def create_visit_viewer(config: dict | None = None):
    """Visit Viewer"""
    from ..widgets.instrument_visits import InstrumentVisits

    root = create_root(window_title='DLS Visits')
    config = config or get_config()

    visit_widgets = []
    beamlines = list(BEAMLINE_CONFIG)
    for row in range(2):
        for col in range(4):
            index = (row * 4) + col
            if index >= len(beamlines):
                break
            beamline = beamlines[index]
            frame = ttk.Frame(root)
            frame.grid(row=row, column=col, padx=5, pady=5)
            visit_widgets.append(InstrumentVisits(frame, beamline, config))

    root.mainloop()
    return root



"""
Crystal Lattice functions for diffraction
"""

import numpy as np

from ..utils.units import unit_converter


def wavevector(wavelength, units='angstrom'):
    """Return wavevector = 2pi/lambda"""
    wavelength_a = unit_converter(wavelength, units, 'angstrom')
    return 2 * np.pi / wavelength_a


def calqmag(twotheta, wavelength, wl_units='angstrom'):
    """
    Calculate |Q| at a particular 2-theta (deg) for energy in keV
     magQ = calqmag(twotheta, wavelength_a=1.5)
    """
    wavelength_a = unit_converter(wavelength, wl_units, 'angstrom')
    theta = twotheta * np.pi / 360  # theta in radians
    magq = np.sin(theta) * 4 * np.pi / wavelength_a
    return magq


def cal2theta(qmag, wavelength, wl_units='angstrom'):
    """
    Calculate 2theta of |Q| in degrees
     twotheta = cal2theta(q_mag, wavelength_a=1.5)
    """
    wavelength_a = unit_converter(wavelength, wl_units, 'angstrom')
    tth = 2 * np.arcsin(qmag * wavelength_a / (4 * np.pi))
    tth = tth * 180 / np.pi
    return tth


def caldspace(twotheta, wavelength, wl_units='angstrom'):
    """
    Calculate d-spacing from two-theta
     dspace = caldspace(tth, wavelength_a=1.5)
    """
    wavelength_a = unit_converter(wavelength, wl_units, 'angstrom')
    qmag = calqmag(twotheta, wavelength_a)
    dspace = q2dspace(qmag)
    return dspace


def q2dspace(qmag):
    """
    Calculate |Q| in inverse Angstroms from d-spacing in Angstroms
         dspace = q2dspace(Qmag)
    """
    return 2 * np.pi / qmag


def dspace2q(dspace):
    """
    Calculate d-spacing in Angstroms from |Q| in inverse Angstroms
         Qmag = q2dspace(dspace)
    """
    return 2 * np.pi / dspace


def bragg_en(energy, d_space, en_units='keV'):
    """Returns the Bragg angle for a given d-space at given photon energy in keV"""
    energy_kev = unit_converter(energy, en_units, 'keV')
    return np.rad2deg(np.arcsin(6.19922 / (energy_kev * d_space)))


def bragg_wl(wavelength, d_space, wl_units='angstrom'):
    """Returns the Bragg angle for a given d-space at given wavelength in Angstroms"""
    wavelength_a = unit_converter(wavelength, wl_units, 'angstrom')
    d_space = np.rad2deg(np.arcsin(wavelength_a / (2 * d_space)))
    return unit_converter(d_space, 'angstrom', wl_units)


def bragg(d_space, wavelength=None, energy=None, wl_units='angstrom', en_units='keV'):
    """Returns the photon Bragg angle in Degrees for a given d-space in Angstroms"""
    if wavelength is None:
        return bragg_en(energy, d_space, en_units=en_units)
    else:
        return bragg_wl(wavelength, d_space, wl_units=wl_units)


def scherrer_size(fwhm, twotheta, wavelength_a, shape_factor=0.9):
    """
    Use the Scherrer equation to calculate the size of a crystallite from a peak width
      L = K * lambda / fwhm * cos(theta)
    See: https://en.wikipedia.org/wiki/Scherrer_equation
    :param fwhm: full-width-half-maximum of a peak, in degrees
    :param twotheta: 2*Bragg angle, in degrees
    :param wavelength_a: incident beam wavelength, in Angstroms
    :param shape_factor: dimensionless shape factor, dependent on shape of crystallite
    :return: float, crystallite domain size in Angstroms
    """

    delta_theta = np.deg2rad(fwhm)
    costheta = np.cos(np.deg2rad(twotheta / 2.))
    return shape_factor * wavelength_a / (delta_theta * costheta)


def scherrer_fwhm(size, twotheta, wavelength_a, shape_factor=0.9):
    """
    Use the Scherrer equation to calculate the size of a crystallite from a peak width
      L = K * lambda / fwhm * cos(theta)
    See: https://en.wikipedia.org/wiki/Scherrer_equation
    :param size: crystallite domain size in Angstroms
    :param twotheta: 2*Bragg angle, in degrees
    :param wavelength_a: incident beam wavelength, in Angstroms
    :param shape_factor: dimensionless shape factor, dependent on shape of crystallite
    :return: float, peak full-width-at-half-max in degrees
    """
    costheta = np.cos(np.deg2rad(twotheta / 2.))
    return np.rad2deg(shape_factor * wavelength_a / (size * costheta))


def bmatrix(a: float, b: float | None = None, c: float | None = None,
            alpha: float = 90., beta: float = 90., gamma: float = 90.) -> np.ndarray:
    """
    Calculate the B matrix as defined in Busing&Levy Acta Cyst. 22, 457 (1967)
    Creates a matrix to transform (hkl) into a cartesian basis:
        (qx,qy,qz)' = B.(h,k,l)'       (where ' indicates a column vector)

    The B matrix is related to the reciprocal basis vectors:
        (astar, bstar, cstar) = 2 * np.pi * B.T
    Where cstar is defined along the z axis

    The B matrix is related to the real-space unit vectors:
        (vec_a, vec_b, vec_c) = B^-1 = inv(B)
    Where vec_b is defined along the y axis

    :param a: lattice parameter a in Angstroms
    :param b: lattice parameter b in Angstroms
    :param c: lattice parameter c in Angstroms
    :param alpha: lattice angle alpha in degrees
    :param beta: lattice angle beta in degrees
    :param gamma: lattice angle gamma in degrees
    :returns: [3x3] array B matrix in inverse-Angstroms (no 2pi)
    """
    if b is None:
        b = a
    if c is None:
        c = a
    alpha1 = np.deg2rad(alpha)
    alpha2 = np.deg2rad(beta)
    alpha3 = np.deg2rad(gamma)

    beta1 = np.arccos((np.cos(alpha2) * np.cos(alpha3) - np.cos(alpha1)) / (np.sin(alpha2) * np.sin(alpha3)))
    beta2 = np.arccos((np.cos(alpha1) * np.cos(alpha3) - np.cos(alpha2)) / (np.sin(alpha1) * np.sin(alpha3)))
    beta3 = np.arccos((np.cos(alpha1) * np.cos(alpha2) - np.cos(alpha3)) / (np.sin(alpha1) * np.sin(alpha2)))

    b1 = 1 / (a * np.sin(alpha2) * np.sin(beta3))
    b2 = 1 / (b * np.sin(alpha3) * np.sin(beta1))
    b3 = 1 / (c * np.sin(alpha1) * np.sin(beta2))

    c1 = b1 * b2 * np.cos(beta3)
    c2 = b1 * b3 * np.cos(beta2)
    c3 = b2 * b3 * np.cos(beta1)

    bm = np.array([
        [b1, b2 * np.cos(beta3), b3 * np.cos(beta2)],
        [0, b2 * np.sin(beta3), -b3 * np.sin(beta2) * np.cos(alpha1)],
        [0, 0, 1 / c]
    ])
    return bm


def reciprocal_lattice(a_vec: np.ndarray, b_vec: np.ndarray, c_vec: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculate the reciprocal lattice vectors from real lattice vectors

    :param a_vec: [3x1] basis vector a in Angstroms
    :param b_vec: [3x1] basis vector b in Angstroms
    :param c_vec: [3x1] basis vector c in Angstroms
    :returns: [3x1], [3x1], [3x1]  reciprocal lattice vectors
    """
    astar, bstar, cstar = 2 * np.pi * np.linalg.inv([a_vec, b_vec, c_vec]).T
    return astar, bstar, cstar




"""
lmfit results interpreter
"""

import numpy as np

from lmfit.model import ModelResult, Model
from matplotlib import pyplot as plt

from mmg_toolbox.utils.misc_functions import stfm
from .models import PEAK_PARS

__all__ = ['peak_results', 'peak_results_str', 'peak_results_fit', 'peak_results_plot', 'Peak', 'FitResults']

class R:
    lmfit = 'lmfit'
    npeaks = 'npeaks'
    peak_prefixes = 'peak_prefixes'
    peak_models = 'peak_models'
    chisqr = 'chisqr'
    xdata = 'xdata'
    ydata = 'ydata'
    weights = 'weights'
    yerror = 'yerror'
    yfit = 'yfit'
    amplitude = 'amplitude'
    center = 'center'
    fwhm = 'fwhm'
    height = 'height'
    sigma = 'sigma'
    background = 'background'
    stderr_amplitude = 'stderr_amplitude'
    stderr_center = 'stderr_center'
    stderr_fwhm = 'stderr_fwhm'
    stderr_height = 'stderr_height'
    stderr_sigma = 'stderr_sigma'
    stderr_background = 'stderr_background'


def peak_results(res: ModelResult) -> dict:
    """
    Generate dict of fit results, including summed totals
    totals = peak_results(res)
    totals = {
        'lmfit': lmfit_result (res),
        'npeaks': number of peak models used,
        'chisqr': Chi^2 of fit,
        'xdata': x-data used for fit,
        'ydata': y-data used for fit,
        'yfit': y-fit values,
        'weights': res.weights,
        'yerror': 1 / res.weights if res.weights is not None else 0 * res.data,
        # plus data from components, e.g.
        'p1_amplitude': Peak 1 area,
        'p1_fwhm': Peak 1 full-width and half-maximum
        'p1_center': Peak 1 peak position
        'p1_height': Peak 1 fitted height,
        # plus data for total fit:
        'amplitude': Total summed area,
        'center': average centre of peaks,
        'height': average height of peaks,
        'fwhm': average FWHM of peaks,
        'background': fitted background,
        # plut the errors on all parameters, e.g.
        'stderr_amplitude': error on 'amplitude',
    }
    :param res: lmfit fit result - ModelResult
    :return: {totals: (value, error)}
    """
    peak_prefx = [mod.prefix for mod in res.model.components if 'bkg' not in mod.prefix]
    peak_models = [mod for mod in res.model.components if 'bkg' not in mod.prefix]
    npeaks = len(peak_prefx)
    nn = 1 / len(peak_prefx) if len(peak_prefx) > 0 else 1  # normalise by number of peaks
    comps = res.eval_components()
    fit_dict = {
        R.lmfit: res,
        R.npeaks: npeaks,
        R.peak_prefixes: peak_prefx,
        R.peak_models: peak_models,
        R.chisqr: res.chisqr,
        R.xdata: res.userkws['x'],
        R.ydata: res.data,
        R.weights: res.weights,
        R.yerror: 1 / res.weights if res.weights is not None else 0 * res.data,
        R.yfit: res.best_fit,
    }
    for comp_prefx, comp in comps.items():
        fit_dict['%sfit' % comp_prefx] = comp
    for pname, param in res.params.items():
        ename = 'stderr_' + pname
        fit_dict[pname] = param.value
        fit_dict[ename] = param.stderr or 0
    totals = {
        R.amplitude: np.sum([res.params['%samplitude' % pfx].value for pfx in peak_prefx]),
        R.center: np.mean([res.params['%scenter' % pfx].value for pfx in peak_prefx]),
        R.sigma: np.mean([res.params['%ssigma' % pfx].value for pfx in peak_prefx]),
        R.height: np.mean([res.params['%sheight' % pfx].value for pfx in peak_prefx]),
        R.fwhm: np.mean([res.params['%sfwhm' % pfx].value for pfx in peak_prefx]),
        R.background: np.mean(comps['bkg_']) if 'bkg_' in comps else 0.0,
        R.stderr_amplitude: np.sqrt(np.sum([fit_dict['stderr_%samplitude' % pfx] ** 2 for pfx in peak_prefx])),
        R.stderr_center: np.sqrt(np.sum([fit_dict['stderr_%scenter' % pfx] ** 2 for pfx in peak_prefx])) * nn,
        R.stderr_sigma: np.sqrt(np.sum([fit_dict['stderr_%ssigma' % pfx] ** 2 for pfx in peak_prefx])) * nn,
        R.stderr_height: np.sqrt(np.sum([fit_dict['stderr_%sheight' % pfx] ** 2 for pfx in peak_prefx])) * nn,
        R.stderr_fwhm: np.sqrt(np.sum([fit_dict['stderr_%sfwhm' % pfx] ** 2 for pfx in peak_prefx])) * nn,
        R.stderr_background: np.std(comps['bkg_']) if 'bkg_' in comps else 0.0,
    }
    fit_dict.update(totals)
    return fit_dict


def peak_results_str(res: ModelResult) -> str:
    """
    Generate output str from lmfit results, including totals
    :param res: lmfit_result
    :return: str
    """
    fit_dict = peak_results(res)
    out = 'Fit Results\n'
    out += '%s\n' % res.model.name
    out += 'Npeaks = %d\n' % fit_dict['npeaks']
    out += 'Method: %s => %s\n' % (res.method, res.message)
    out += 'Chisqr = %1.5g\n' % res.chisqr
    # Peaks
    peak_prefx = [mod.prefix for mod in res.model.components if 'bkg' not in mod.prefix]
    for prefx in peak_prefx:
        out += '\nPeak %s\n' % prefx
        for pn in res.params:
            if prefx in pn:
                out += '%15s = %s\n' % (pn, stfm(fit_dict[pn], fit_dict['stderr_%s' % pn]))

    out += '\nBackground\n'
    for pn in res.params:
        if 'bkg' in pn:
            out += '%15s = %s\n' % (pn, stfm(fit_dict[pn], fit_dict['stderr_%s' % pn]))

    out += '\nTotals\n'
    out += '      amplitude = %s\n' % stfm(fit_dict[R.amplitude], fit_dict[R.stderr_amplitude])
    out += '         center = %s\n' % stfm(fit_dict[R.center], fit_dict[R.stderr_center])
    out += '         height = %s\n' % stfm(fit_dict[R.height], fit_dict[R.stderr_height])
    out += '          sigma = %s\n' % stfm(fit_dict[R.sigma], fit_dict[R.stderr_sigma])
    out += '           fwhm = %s\n' % stfm(fit_dict[R.fwhm], fit_dict[R.stderr_fwhm])
    out += '     background = %s\n' % stfm(fit_dict[R.background], fit_dict[R.stderr_background])
    return out


def peak_results_fit(res: ModelResult, ntimes: int = 10, x_data: np.ndarray | None = None) -> tuple[np.ndarray, np.ndarray]:
    """
    Generate xfit, yfit data, interpolated to give smoother variation
    :param res: lmfit_result
    :param ntimes: int, number of points * old number of points
    :param x_data: x data to interpolate or None to use data from fit
    :return: xfit, yfit
    """
    old_x = res.userkws['x'] if x_data is None else x_data
    xfit = np.linspace(np.min(old_x), np.max(old_x), np.size(old_x) * ntimes)
    yfit = res.eval(x=xfit)
    return xfit, yfit


def peak_results_plot(res: ModelResult, axes=None, xlabel: str = None, ylabel: str = None, title: str = None):
    """
    Plot peak results
    :param res: lmfit result
    :param axes: None or matplotlib axes
    :param xlabel: None or str
    :param ylabel: None or str
    :param title: None or str
    :return: matplotlib figure or axes
    """
    xdata = res.userkws['x']
    if title is None:
        title = res.model.name

    if axes:
        ax = res.plot_fit(ax=axes, xlabel=xlabel, ylabel=ylabel)
        # Add peak components
        comps = res.eval_components(x=xdata)
        for component in comps.keys():
            ax.plot(xdata, comps[component], label=component)
            ax.legend()
        return ax

    fig = res.plot(xlabel=xlabel, ylabel=ylabel)
    try:
        fig, grid = fig  # Old version of LMFit
    except TypeError:
        pass

    ax1, ax2 = fig.axes
    ax1.set_title(title, wrap=True)
    # Add peak components
    comps = res.eval_components(x=xdata)
    for component in comps.keys():
        ax2.plot(xdata, comps[component], label=component)
        ax2.legend()
    fig.set_figwidth(8)
    # fig.show()
    return fig


class Peak:
    """
    Peak object
    """

    def __init__(self, result: ModelResult, model: Model, amplitude: float, center: float, height: float, fwhm: float,
                 stderr_amplitude: float, stderr_center: float, stderr_height: float, stderr_fwhm: float, **kwargs):
        self._result = result
        self.model = model
        self.prefix = model.prefix
        self.model_name = model._name
        self.params = PEAK_PARS
        self.amplitude = amplitude
        self.center = center
        self.height = height
        self.fwhm = fwhm
        self.stderr_amplitude = stderr_amplitude
        self.stderr_center = stderr_center
        self.stderr_height = stderr_height
        self.stderr_fwhm = stderr_fwhm
        for name, value in kwargs.items():
            setattr(self, name, value)
            if name.startswith('stderr_') and hasattr(self, name[7:]):
                self.params.append(name[7:])

    def __repr__(self):
        pars = ', '.join(f"{p}={self.get_string(p)}" for p in self.params)
        return f"Peak('{self.prefix}', '{self.model_name}', {pars})"

    def get_value(self, name: str) -> tuple[float | None, float]:
        """Returns fit parameter value and associated error"""
        err_name = f"stderr_{name}"
        if not hasattr(self, err_name):
            return None, 0
        value = getattr(self, name)
        error = getattr(self, err_name)
        return value, error

    def get_string(self, name: str) -> str:
        """Returns fit parameter string including error in standard form"""
        value, error = self.get_value(name)
        if value is None:
            return '--'
        return stfm(value, error)

    def fit_data(self, x_data: np.ndarray | None = None, ntimes=10) -> tuple[np.ndarray, np.ndarray]:
        """Returns interpolated x, y fit arrays"""
        old_x = self._result.userkws['x'] if x_data is None else x_data
        xfit = np.linspace(np.min(old_x), np.max(old_x), np.size(old_x) * ntimes)
        yfit = self.model.eval(x=xfit)
        return xfit, yfit

    def label(self) -> str:
        return f"{self.model_name} ({self.prefix})"

    def plot(self, axes: plt.Axes, x_data: np.ndarray | None = None, ntimes=10):
        """Plot peak fit results"""
        xfit, yfit = self.fit_data(x_data, ntimes)
        axes.plot(xfit, yfit, label=self.label())


class FitResults:
    """
    FitResults Class
    Wrapper for lmfit results object with additional functions specific to i16_peakfit

    res = model.fit(ydata, x=xdata)  # lmfit ouput
    fitres = FitResults(res)

    --- Parameters ---
    fitres.res  # lmfit output
    # data from fit:
    fitres.npeaks # number of peak models used,
    fitres.chisqr  # Chi^2 of fit,
    fitres.xdata  # x-data used for fit,
    fitres.ydata  # y-data used for fit,
    fitres.yfit  # y-fit values,
    fitres.weights  # res.weights,
    fitres.yerror  # 1 / res.weights if res.weights is not None else 0 * res.data,
    # data from components, e.g.
    fitres.p1_amplitude  # Peak 1 area,
    fitres.p1_fwhm  # Peak 1 full-width and half-maximum
    fitres.p1_center  # Peak 1 peak position
    fitres.p1_height  # Peak 1 fitted height,
    # data for total fit:
    fitres.amplitude  # Total summed area,
    fitres.center  # average centre of peaks,
    fitres.height  # average height of peaks,
    fitres.fwhm  # average FWHM of peaks,
    fitres.background  # fitted background,
    # errors on all parameters, e.g.
    fitres.stderr_amplitude  # error on 'amplitude

    --- Functions ---
    print(fitres)  # prints formatted str with results
    ouputdict = fitres.results()  # creates ouput dict
    xdata, yfit = fitres.fit(ntimes=10)  # interpolated fit results
    fig = fitres.plot(axes, xlabel, ylabel, title)  # create plot
    """
    npeaks: int
    peak_prefixes: list[str]
    peak_models: list[Model]
    amplitude: float
    center: float
    height: float
    fwhm: float
    background: float
    stderr_amplitude: float
    stderr_center: float
    stderr_height: float
    stderr_fwhm: float
    stderr_background: float
    chisqr: float  # Chi^2 of fit,
    xdata: np.ndarray  # x-data used for fit,
    ydata: np.ndarray  # y-data used for fit,
    yfit: np.ndarray  # y-fit values,
    weights: np.ndarray  # res.weights,
    yerror: np.ndarray  # 1 / res.weights if res.weights is not None else 0 * res.data,

    def __init__(self, results: ModelResult):
        self.res = results
        self._res = peak_results(results)
        self.params = PEAK_PARS
        for name, value in self._res.items():
            setattr(self, name, value)

    def __repr__(self):
        pars = ', '.join(f"{p}={self.get_string(p)}" for p in self.params)
        return f"FitResults(npeaks={self.npeaks}, {pars})"

    def __str__(self):
        return peak_results_str(self.res)

    def __getitem__(self, item: int | slice) -> Peak | list[Peak]:
        if isinstance(item, slice):
            return [self.get_peak(n) for n in range(*item.indices(len(self)))]
        else:
            return self.get_peak(item)

    def __len__(self):
        return self.npeaks

    def peak_name(self, number) -> tuple[str, str]:
        prefix = self.peak_prefixes[number]
        return prefix, "stderr_" + prefix

    def get_peak(self, number: int) -> Peak:
        if number >= self.npeaks:
            raise IndexError('Not enough peaks')
        prefix, stderr = self.peak_name(number)
        pars = {p: self._res.get(prefix + p, 0) for p in self.params}
        stderr = {f"stderr_{p}": self._res.get(stderr + p, 0) for p in self.params}
        return Peak(
            result=self.res,
            model=self.peak_models[number],
            **pars,
            **stderr
        )

    def get_value(self, name: str) -> tuple[float | None, float]:
        """Returns fit parameter value and associated error"""
        err_name = f"stderr_{name}"
        value = self._res.get(name, None)
        error = self._res.get(err_name, 0)
        return value, error

    def get_string(self, name: str) -> str:
        """Returns fit parameter string including error in standard form"""
        value, error = self.get_value(name)
        return stfm(value, error)

    def results(self) -> dict:
        """Returns dict of peak fit results"""
        return self._res

    def fit_data(self, x_data: np.ndarray | None = None, ntimes=10) -> tuple[np.ndarray, np.ndarray]:
        """Returns interpolated x, y fit arrays"""
        return peak_results_fit(self.res, ntimes=ntimes, x_data=x_data)

    def plot(self, axes=None, xlabel=None, ylabel=None, title=None):
        """Plot peak fit results"""
        return peak_results_plot(self.res, axes, xlabel, ylabel, title)




"""
X-Ray Absorption Spectroscopy (XAS) tools
"""

from .spectra_analysis import xray_edges_in_range, energy_range_edge_label
from .spectra import Spectra, SpectraSubtraction, SpectraAverage
from .spectra_container import SpectraContainer, SpectraContainerSubtraction
from .nxxas_loader import load_xas_scans, create_xas_scan, find_similar_measurements

__all__ = [
    'Spectra', 'SpectraSubtraction', 'SpectraAverage', 'SpectraContainer', 'SpectraContainerSubtraction',
    'load_xas_scans', 'create_xas_scan', 'find_similar_measurements',
    'xray_edges_in_range', 'energy_range_edge_label'
]



"""
Beamline or User Configuration Options
"""

from mmg_toolbox.utils.env_functions import YEAR
from .hdfmap_generic import HdfMapMMGMetadata as Md
from .metadata_strings import BEAMLINE_META, META_STRING, META_LABEL


class C:
    """Names used in config object"""
    conf_file = 'config_file'
    default_directory = 'default_directory'
    normalise_factor = 'normalise_factor'
    replace_names = 'replace_names'
    metadata_string = 'metadata_string'
    metadata_list = 'metadata_list'
    metadata_label = 'metadata_label'
    default_metadata = 'default_metadata'
    beamline = 'beamline'
    scan_description = 'scan_description'
    roi = 'roi'

DEFAULT_SCAN_DESCRIPTION = '{(cmd|user_command|scan_command)}'

META_LIST = {
    # scan number and start_time included by default
    # name: format
    'cmd': DEFAULT_SCAN_DESCRIPTION
}

REPLACE_NAMES = {
    # NEW_NAME: EXPRESSION
    '_t': '(count_time|counttime|t?(1.0))',
}

ROIs: list[tuple[str, str | int, str | int, int, int, str]] = [
    # (name, cen_i, cen_j, wid_i, wid_j, det_name)
]

CONFIG = {
    C.conf_file: '',
    C.normalise_factor: '',
    C.replace_names: REPLACE_NAMES,
    C.roi: ROIs,
    C.metadata_string: META_STRING,
    C.metadata_list: META_LIST,
    C.metadata_label: META_LABEL,
    C.default_metadata: Md.temp,
    C.scan_description: DEFAULT_SCAN_DESCRIPTION,
}

BEAMLINE_CONFIG = {
    'i06': {
        C.beamline: 'i06',
        C.default_directory: f"/dls/i06/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i06'],
        C.normalise_factor: '',
        # C.default_metadata: '',
    },
    'i06-1': {
        C.beamline: 'i06-1',
        C.default_directory: f"/dls/i06-1/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i06-1'],
        C.normalise_factor: '',
        # C.default_metadata: '',
    },
    'i06-2': {
        C.beamline: 'i06-2',
        C.default_directory: f"/dls/i06-2/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i06-2'],
        C.normalise_factor: '',
        # C.default_metadata: '',
    },
    'i10': {
        C.beamline: 'i10',
        C.default_directory: f"/dls/i10/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i10'],
        C.normalise_factor: '',
        # C.default_metadata: '',
    },
    'i10-1': {
        C.beamline: 'i10-1',
        C.default_directory: f"/dls/i10-1/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i10-1'],
        C.normalise_factor: '/(mcs16|macr16|mcse16|macj316|mcsh16|macj216)',
        # C.default_metadata: '',
    },
    'i16': {
        C.beamline: 'i16',
        C.default_directory: f"/dls/i16/data/{YEAR}/",
        C.normalise_factor: '/Transmission/count_time/(rc/300.)',
        C.metadata_string: BEAMLINE_META['i16'],
        C.roi: [
            ('pilroi1', 'pil3_centre_j', 'pil3_centre_i', 30, 30, 'pil3_100k'),
        ],
        C.default_metadata: 'Tsample',
    },
    'i21': {
        C.beamline: 'i21',
        C.default_directory: f"/dls/i21/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i21'],
        C.normalise_factor: '',
        # C.default_metadata: '',
    },
}


def beamline_config(beamline: str | None = None) -> dict:
    """Returns the default beamline config dict"""
    config = CONFIG.copy()
    if beamline and beamline in BEAMLINE_CONFIG:
        config.update(BEAMLINE_CONFIG[beamline])
    elif beamline:
        config[C.beamline] = beamline
    return config




"""
mmg_toolbox tests
Test polarisation functions
"""

import pytest
import numpy as np
import hdfmap

from mmg_toolbox.utils import polarisation as pol
from . import only_dls_file_system
from .example_files import FILES_DICT


def test_polarisation():
    assert pol.check_polarisation(pol.PolLabels.linear_horizontal) == 'lh'
    assert pol.check_polarisation(pol.PolLabels.linear_vertical) == 'lv'
    assert pol.check_polarisation(pol.PolLabels.circular_left) == 'cl'
    assert pol.check_polarisation(pol.PolLabels.circular_right) == 'cr'
    assert pol.check_polarisation(pol.PolLabels.circular_negative) == 'cl'
    assert pol.check_polarisation(pol.PolLabels.circular_positive) == 'cr'
    assert pol.check_polarisation(pol.PolLabels.linear_arbitrary, 0) == 'lh'
    assert pol.check_polarisation(pol.PolLabels.linear_arbitrary, 90) == 'lv'
    assert pol.check_polarisation(pol.PolLabels.linear_arbitrary, 30) == 'la'
    assert pol.check_polarisation(np.array([1,1,0,0])) == 'lh'
    assert pol.check_polarisation(None, 60) == 'la'
    assert pol.pol_subtraction_label('pc') == 'xmcd'
    assert pol.pol_subtraction_label('lv') == 'xmld'


@only_dls_file_system
def test_read_polarisation():
    filename = FILES_DICT['i16 pilatus eta scan, new nexus format']
    with hdfmap.load_hdf(filename) as hdf:
        polarisation = pol.get_polarisation(hdf)

    assert polarisation == pol.PolLabels.linear_horizontal

    filename = FILES_DICT['i10-1 la pol']

    with hdfmap.load_hdf(filename) as hdf:
        polarisation = pol.get_polarisation(hdf)

    assert polarisation == pol.PolLabels.linear_arbitrary


def test_arbitrary_polarisation():
    p0, p1, p2, p3 = pol.stokes_from_vector(30)
    assert pol.polarisation_label_from_stokes(p0, p1, p2, p3) == 'la'


"""
{{title}}
{{description}}
"""

import matplotlib.pyplot as plt
import hdfmap

filenames = [
    {{filepaths}}
]

fig, ax = plt.subplots()

for file in filenames:
    m = hdfmap.NexusMap()
    with hdfmap.load_hdf(file) as nxs:
        m.populate(nxs)
        data = m.get_plot_data(nxs)
        label = m.format_hdf(nxs, "{filename}: {__axes}")
    ax.plot(data['xdata'], data['ydata'], label=label)

ax.set_xlabel(data['xlabel'])
ax.set_ylabel(data['ylabel'])
ax.set_title(data['title'])

plt.show()



"""
tk window to select HDF paths from files
"""

import tkinter as tk
from tkinter import ttk
import hdfmap

from ..misc.functions import open_close_all_tree
from ..misc.config import get_config
from ..misc.styles import create_root


def create_metadata_selector(hdf_map: hdfmap.NexusMap,
                             parent: tk.Misc | None = None, config: dict | None = None) -> list[str]:
    """
    Create a hdfmap namespace selector
    """
    from ..widgets.nexus_treeview import HdfNameSpace

    root = create_root(parent=parent, window_title='Select Metadata')
    config = get_config() if config is None else config

    widget = HdfNameSpace(root)
    with hdf_map.load_hdf() as hdf:
        widget.populate(hdf, hdf_map, all=False, metadata=True)
    open_close_all_tree(widget.tree, "", True)

    output_names = []

    def select():
        output_names.extend([
            widget.tree.item(iid, 'text')
            for iid in widget.tree.selection()
        ])
        root.destroy()

    ttk.Button(root, text='Select', command=select).pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=5)

    root.wait_window()
    return output_names


def create_scannable_selector(hdf_map: hdfmap.NexusMap,
                              parent: tk.Misc | None = None, config: dict | None = None) -> list[str]:
    """
    Create a hdfmap namespace selector
    """
    from ..widgets.nexus_treeview import HdfNameSpace

    root = create_root(parent=parent, window_title='Select Scannable')
    config = get_config() if config is None else config

    widget = HdfNameSpace(root)
    with hdf_map.load_hdf() as hdf:
        widget.populate(hdf, hdf_map, all=False, scannables=True)
    open_close_all_tree(widget.tree, "", True)

    output_names = []

    def select(event=None):
        output_names.extend([
            widget.tree.item(iid, 'text')
            for iid in widget.tree.selection()
        ])
        root.destroy()

    widget.tree.bind("<Double-1>", select)
    ttk.Button(root, text='Select', command=select).pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=5)

    root.wait_window()
    return output_names


"""
a tkinter frame with a single plot
"""
import os
import tkinter as tk
from tkinter import ttk
import numpy as np

import hdfmap
from hdfmap import create_nexus_map

from mmg_toolbox.utils.file_functions import read_tiff, get_scan_number
from ..misc.styles import create_hover, create_root
from ..misc.screen_size import get_figure_size
from ..misc.matplotlib import ini_image, COLORMAPS, DEFAULT_COLORMAP, add_rectangle
from ..misc.logging import create_logger
from ..misc.config import get_config, C
from .roi_editor import RoiEditor

logger = create_logger(__file__)


class NexusDetectorImage:
    def __init__(self, root: tk.Misc, hdf_filename: str | None = None,
                 config: dict | None = None, hdf_map: hdfmap.NexusMap | None = None):
        self.filename = hdf_filename
        self.parent = root
        self.map = hdf_map
        self.config = config or get_config()

        self.detector_name = tk.StringVar(root, 'NXdetector')
        self.view_index = tk.IntVar(root, 0)
        self.axis_name = tk.StringVar(root, 'axis = ')
        self.axis_value = tk.DoubleVar(root, 0)
        self.logplot = tk.BooleanVar(root, False)
        self.difplot = tk.BooleanVar(root, False)
        self.mask = tk.DoubleVar(root, 0)
        self.cmin = tk.DoubleVar(root, 0)
        self.cmax = tk.DoubleVar(root, 1)
        self.fixclim = tk.BooleanVar(root, False)
        self.colormap = tk.StringVar(root, self.config.get(C.default_colormap, DEFAULT_COLORMAP))
        self.image_error = tk.StringVar(root, '')
        self.extra_plot_callbacks = []  # calls any function in this list on update_image
        self.roi_names = []

        section = ttk.Frame(root)
        section.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        frm = ttk.Frame(section)
        frm.pack(side=tk.TOP, expand=tk.NO, fill=tk.X)
        ttk.Button(frm, text='Window', command=self.new_window).pack(side=tk.LEFT, padx=5)
        ttk.Button(frm, text='ROIs', command=self.roi_frame).pack(side=tk.LEFT)
        ttk.Button(frm, text='Draw new ROI', command=self.mouse_select_roi).pack(side=tk.LEFT)
        self.detector_menu = ttk.OptionMenu(frm, self.detector_name, None, 'NXdetector', command=self.update_image_plot)
        self.detector_menu.pack(side=tk.RIGHT)

        frm = ttk.Frame(section)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        self.im_fig, self.im_ax, self._im_lines, self.ax_image, self.colorbar, self.toolbar = ini_image(
            frame=frm,
            figure_size=get_figure_size(root, self.config, C.image_size),
            figure_dpi=self.config.get(C.plot_dpi),
        )
        self.im_ax.set_xlabel(None)
        self.im_ax.set_ylabel(None)
        self.rectangle = self.im_ax.axvspan(0, 0, alpha=0.2)

        # Error message
        frm = ttk.Frame(section)
        frm.pack(side=tk.TOP, expand=tk.NO, fill=tk.X)
        self.error_label = ttk.Label(frm, textvariable=self.image_error, style='error.TLabel')  # only pack this on error

        self.slider = self.ini_slider(root)

        if hdf_filename:
            self.update_image_data_from_file(hdf_filename, hdf_map)

    def ini_slider(self, root: tk.Misc):
        frm = ttk.Frame(root)
        frm.pack(expand=tk.NO, pady=2, padx=5)

        def inc():
            new_val = self.view_index.get() + 1
            if new_val < self.map.scannables_length():
                self.view_index.set(new_val)
                self.update_image()

        def dec():
            new_val = self.view_index.get() - 1
            if new_val >= 0:
                self.view_index.set(new_val)
                self.update_image()

        var = ttk.Label(frm, text='Index:', width=8)
        var.pack(side=tk.LEFT)
        var = ttk.Button(frm, text='-', command=dec)
        var.pack(side=tk.LEFT)
        tkscale = ttk.Scale(frm, from_=0, to=100, variable=self.view_index, orient=tk.HORIZONTAL,
                            command=self.update_image, length=300)
        # var.bind("<ButtonRelease-1>", callback)
        tkscale.pack(side=tk.LEFT, expand=tk.YES)
        var = ttk.Button(frm, text='+', command=inc)
        var.pack(side=tk.LEFT)
        var = ttk.Entry(frm, textvariable=self.view_index, width=6)
        var.pack(side=tk.LEFT)
        var.bind('<Return>', self.update_image)
        var.bind('<KP_Enter>', self.update_image)

        # axis mode
        var = ttk.Label(frm, textvariable=self.axis_name)
        var.pack(side=tk.LEFT)
        var = ttk.Label(frm, textvariable=self.axis_value)
        var.pack(side=tk.LEFT)

        # Options button
        var = ttk.Button(frm, text='Options', command=self.options_frame)
        var.pack(side=tk.LEFT, padx=3)
        return tkscale

    def options_frame(self):
        """A hovering frame with options"""
        window, fun_close = create_hover(self.parent)

        frm = ttk.LabelFrame(window, text='Options', relief=tk.RIDGE)
        frm.pack(expand=tk.NO, pady=2, padx=5)

        var = ttk.Checkbutton(frm, text='Log', variable=self.logplot, command=self.update_image)
        var.pack(side=tk.LEFT, padx=6)
        var = ttk.Checkbutton(frm, text='Diff', variable=self.difplot, command=self.update_image)
        var.pack(side=tk.LEFT, padx=6)

        var = ttk.Label(frm, text='Mask <')
        var.pack(side=tk.LEFT, expand=tk.NO, padx=6)
        var = ttk.Entry(frm, textvariable=self.mask, width=6)
        var.pack(side=tk.LEFT, padx=6)
        var.bind('<Return>', self.update_image)
        var.bind('<KP_Enter>', self.update_image)

        var = ttk.OptionMenu(frm, self.colormap, self.colormap.get(), *COLORMAPS,
                             command=self.update_colormap_details)
        var.pack(side=tk.LEFT)

        var = ttk.Label(frm, text='clim:')
        var.pack(side=tk.LEFT, expand=tk.NO)
        var = ttk.Entry(frm, textvariable=self.cmin, width=6)
        var.pack(side=tk.LEFT)
        var.bind('<Return>', self.update_colormap_details)
        var.bind('<KP_Enter>', self.update_colormap_details)
        var = ttk.Entry(frm, textvariable=self.cmax, width=6)
        var.pack(side=tk.LEFT)
        var.bind('<Return>', self.update_colormap_details)
        var.bind('<KP_Enter>', self.update_colormap_details)
        var = ttk.Checkbutton(frm, text='Fix', variable=self.fixclim)
        var.pack(side=tk.LEFT)

        frm = ttk.Frame(window)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.X)
        var = ttk.Button(frm, text='Close', command=fun_close)
        var.pack(fill=tk.X)

    def roi_frame(self):
        """a hovering frame with ROIs"""
        # window, fun_close = create_hover(root)
        window = create_root('Regions of Interest (ROIs)', self.parent)

        def on_close():
            window.destroy()
            self.add_config_rois()
            self.plot_config_rois()

        RoiEditor(window, self.config, on_close)

    def options_menu(self) -> dict:
        """Return a dict of options for menu structure"""
        menu = {
            "Image": {
                "Change options": self.options_frame,
            }
        }
        return menu

    def new_window(self):
        window = create_root(self.filename, self.parent)
        widget = NexusDetectorImage(window, self.filename, self.config, self.map)
        return widget

    def _clear_image_error(self):
        self.error_label.pack_forget()

    def _show_image_error(self, message):
        self.image_error.set(message)
        self.error_label.pack()

    def im_line(self, *args, **kwargs):
        lines = self.im_ax.plot(*args, **kwargs)
        self._im_lines.extend(lines)

    def text(self, *args, **kwargs):
        text = self.im_ax.text(*args, **kwargs)
        self._im_lines.append(text)

    def remove_im_lines(self):
        for obj in self._im_lines:
            obj.remove()
        self._im_lines.clear()

    def update_image_data_from_file(self, filename: str, hdf_map: hdfmap.NexusMap | None = None):
        logger.debug(f'update image from filename: {filename}')
        self._clear_image_error()
        self.filename = filename
        self.map = create_nexus_map(self.filename) if hdf_map is None else hdf_map
        self.slider.config(to=self.map.scannables_length() - 1)  # set slider max
        detector_names = list(self.map.image_data.keys())
        # self.detector_menu['values'] = detector_names if detector_names else ['No 2D NXdetectors']
        self.detector_menu.set_menu(
            *detector_names
        )
        if not detector_names:
            return
        # TODO add axis_name and value
        self.add_config_rois()
        self.view_index.set(0)
        self.update_image_plot()

    def _get_image(self):
        try:
            detector = self.detector_name.get()
            axis_name = self.axis_name.get()
            index = int(self.view_index.get())
            self.view_index.set(index)
            logger.debug(f"load image: {detector} [{index}] with axis '{axis_name}'")

            self.map.set_image_path(self.map.image_data[detector])
            with hdfmap.load_hdf(self.filename) as hdf:
                image = self.map.get_image(hdf, index)
                value = self.map.get_data(hdf, axis_name, index=index, default=index)

            if issubclass(type(image), str):
                # TIFF image, NXdetector/image_data -> array('file.tif')
                file_directory = os.path.dirname(self.filename)
                image_filename = os.path.join(file_directory, image)
                # print(f"directory: {file_directory}\nstring: {image}\nfilename: {image_filename}")
                logger.info(f"load tiff image from '{image}': {image_filename}")
                if not os.path.isfile(image_filename):
                    raise FileNotFoundError(f"File not found: {image_filename}")
                image = read_tiff(image_filename)
            elif image.ndim == 0:
                # image is file path number, NXdetector/path -> arange(n_points)
                scan_number = get_scan_number(self.filename)
                file_directory = os.path.dirname(self.filename)
                image_filename = os.path.join(file_directory, f"{scan_number}-{detector}-files/{image:05.0f}.tif")
                logger.info(f"load tiff image from {image}: {image_filename}")
                if not os.path.isfile(image_filename):
                    raise FileNotFoundError(f"File not found: {image_filename}")
                image = read_tiff(image_filename)
            elif image.ndim != 2:
                raise Exception(f"detector image[{index}] is the wrong shape: {image.shape}")
        except Exception as e:
            self._show_image_error(f'Error loading image: {e}')
            image = np.zeros([10, 10])
            value = np.nan
        return image, value

    def update_image_plot(self, event=None):
        """replace plot instance (e.g. on loading new file)"""
        self._clear_image_error()
        if self.filename is None:
            return
        image, value = self._get_image()

        # Options
        if self.logplot.get():
            image = np.log10(image)
        if self.difplot.get():
            raise Warning('Not implemented yet')
        if self.mask.get():
            raise Warning('Not implemented yet')

        cmap = self.colormap.get()
        cmin, cmax = self._get_clim()
        if not self.fixclim.get():
            cmax = 1 + image.max() / 2
            self.cmax.set(cmax)
        # Add plot by removing old one
        self.ax_image.remove()
        self.rectangle.remove()
        self.ax_image = self.im_ax.pcolormesh(image, shading='auto', clim=[cmin, cmax], cmap=cmap)
        self.im_ax.set_xlim([0, image.shape[1]])
        self.im_ax.set_ylim([0, image.shape[0]])
        # self.rectangle = self.ax.axvspan(0, 0, alpha=0.9, facecolor='k', zorder=2)
        self.rectangle = add_rectangle(self.im_ax, 0, 0, 0, 0)
        self.plot_config_rois()
        self.colorbar.update_normal(self.ax_image)
        self.toolbar.update()
        self.im_fig.canvas.draw()
        # Load axis mode
        self.axis_value.set(value)
        # Run additional callbacks
        for function in self.extra_plot_callbacks:
            function()

    def update_image(self, event=None):
        """replace array in plot (e.g. on changing slider)"""
        self._clear_image_error()
        if self.filename is None:
            return
        image, value = self._get_image()

        if self.logplot.get():
            image = np.log10(image)

        self.ax_image.set_array(image)
        self.axis_value.set(value)
        self.update_colormap_details()
        # Run additional callbacks
        for function in self.extra_plot_callbacks:
            function()

    def _get_clim(self):
        cmin, cmax = self.cmin.get(), self.cmax.get()
        if self.logplot.get():
            cmin, cmax = np.log10([cmin + 1, cmax])
        return cmin, cmax

    def update_colormap_details(self, event=None):
        """Update plot colormap details (e.g. on changing colormap)"""
        cmap = self.colormap.get()
        self.ax_image.set_clim(self._get_clim())
        self.ax_image.set_cmap(cmap)
        self.colorbar.update_normal(self.ax_image)
        self.toolbar.update()
        self.im_fig.canvas.draw()

    def add_roi(self, name: str, cen_i: int | str, cen_j: int | str,
                wid_i: int = 30, wid_j: int = 30, image_name: str = 'IMAGE'):
        """add region of interest to config"""
        if C.roi in self.config:
            self.config[C.roi].append((name, cen_i, cen_j, wid_i, wid_j, image_name))
        self.add_config_rois()
        self.plot_config_rois()
        print(f"ROI created: {name}, {cen_i}, {cen_j}, {wid_i}, {wid_j}")

    def add_config_rois(self):
        """add config rois to hdfmap"""
        self.roi_names.clear()
        rois = self.config.get(C.roi)
        detector_names = list(self.map.image_data.keys())
        if rois:
            for name, cen_i, cen_j, wid_i, wid_j, det_name in rois:
                if det_name in detector_names:
                    self.map.add_roi(name, cen_i, cen_j, wid_i, wid_j, det_name)
                    self.roi_names.extend([
                        f"{name}_total",
                        f"{name}_max",
                        f"{name}_min",
                        f"{name}_mean",
                        f"{name}_bkg",
                        f"{name}_rmbkg",
                        # f"{name}_box",
                        # f"{name}_bkg_box",
                    ])

    def plot_config_rois(self):
        """plot config rois on image"""
        self.remove_im_lines()
        rois = self.config.get(C.roi)
        detector = self.detector_name.get()
        if rois:
            #TODO: replace with roi_box from hdfmap
            try:
                with hdfmap.load_hdf(self.filename) as hdf:
                    for n, (name, cen_i, cen_j, wid_i, wid_j, det_name) in enumerate(rois):
                        if det_name == detector:
                            cen_i, cen_j, wid_i, wid_j = self.map.eval(hdf, f"{cen_i},{cen_j},{wid_i},{wid_j}")
                            roi_square = np.array([
                                # x, y
                                [cen_i - wid_i // 2, cen_j - wid_j // 2],
                                [cen_i - wid_i // 2, cen_j + wid_j // 2],
                                [cen_i + wid_i // 2, cen_j + wid_j // 2],
                                [cen_i + wid_i // 2, cen_j - wid_j // 2],
                                [cen_i - wid_i // 2, cen_j - wid_j // 2],
                            ])
                            self.im_line(roi_square[:, 1], roi_square[:, 0], 'k-', lw=2)
                            self.text(cen_j + wid_j // 2, cen_i + wid_i // 2, str(n))
            except Exception as e:
                self._show_image_error(f'Error plotting ROIs: {e}')
        self.im_fig.canvas.draw()

    def mouse_select_roi(self):
        """Select a roi with the mouse"""
        x_start = [0]
        y_start = [0]
        ipress = [False]

        def disconnect():
            self.im_fig.canvas.mpl_disconnect(press)
            self.im_fig.canvas.mpl_disconnect(move)
            self.im_fig.canvas.mpl_disconnect(release)
            self.im_fig.canvas._tkcanvas.master.config(cursor="arrow")

        def mouse_press(event):
            if event.inaxes:
                x_start[0] = event.xdata
                y_start[0] = event.ydata
                ipress[0] = True
            else:
                disconnect()

        def mouse_move(event):
            if event.inaxes and ipress[0]:
                x_end = event.xdata
                y_end = event.ydata
                self.rectangle.set_bounds(x_start[0], y_start[0], x_end - x_start[0], y_end - y_start[0])
                self.im_fig.canvas.draw()

        def mouse_release(event):
            x_end = event.xdata
            y_end = event.ydata
            x_wid = x_end - x_start[0]
            y_wid = y_end - y_start[0]
            x_cen = x_start[0] + x_wid/2
            y_cen = y_start[0] + y_wid/2
            detector = self.detector_name.get()
            name = f"{detector}_roi"
            roi_count = 1
            roi_names = [roi[0] for roi in self.config.get(C.roi)]
            while f"{name}{roi_count}" in roi_names:
                roi_count += 1
            self.add_roi(
                name=f"{name}{roi_count}",
                cen_i=int(y_cen),
                cen_j=int(x_cen),
                wid_i=int(y_wid),
                wid_j=int(x_wid),
                image_name=detector
            )
            self.rectangle.set_bounds(0, 0, 0, 0)
            self.im_fig.canvas.draw()
            disconnect()

        press = self.im_fig.canvas.mpl_connect('button_press_event', mouse_press)
        move = self.im_fig.canvas.mpl_connect('motion_notify_event', mouse_move)
        release = self.im_fig.canvas.mpl_connect('button_release_event', mouse_release)
        self.im_fig.canvas._tkcanvas.master.config(cursor="crosshair")



"""
tk widget for editing the Config file
"""

from ..misc.styles import tk, ttk
from ..misc.logging import create_logger
from ..misc.config import get_config, C

logger = create_logger(__file__)


class RoiEditor:
    """
    Edit the Configuration File in an inset window
    """

    def __init__(self, root: tk.Misc, config: dict | None = None, close_func = None):
        self.root = root
        self.close_func = close_func
        if config is None:
            self.config = get_config()
        else:
            self.config = config

        window = ttk.Frame(root, borderwidth=2, relief=tk.RIDGE)
        window.pack(expand=tk.NO, pady=2, padx=5)

        self.roi_table = ttk.Frame(window, borderwidth=20, relief=tk.RAISED)
        self.roi_table.pack(side=tk.TOP, fill=tk.BOTH)

        frm = ttk.Frame(self.roi_table)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
        ttk.Label(frm, text='Image ROIs', style="title.TLabel").pack(expand=tk.YES, fill=tk.X, padx=10, pady=10)

        self.roi_values: list[tuple] = []
        self.roi_lines: list[tuple[bool, ttk.Frame]] = []
        rois = self.config.get(C.roi)
        if rois:
            for name, cen_i, cen_j, wid_i, wid_j, det_name in rois:
                self.add_roi(name, cen_i, cen_j, wid_i, wid_j, det_name)

        ln = ttk.Frame(window, borderwidth=2)
        ln.pack(side=tk.TOP, fill=tk.X)
        ttk.Button(ln, text='Add ROI', command=self.add_roi).pack()

        ln = ttk.Frame(window, borderwidth=6)
        ln.pack(side=tk.BOTTOM, fill=tk.X)
        ttk.Button(ln, text='Close', command=self.close).pack(fill=tk.X, expand=tk.YES)

    def add_roi(self, name: str = '', cen_i: int | str = '', cen_j: int | str = '',
                wid_i: int = 30, wid_j: int = 30, det_name: str = 'IMAGE'):
        tkvars = (
            # label, tkVar, width
            ('Name:', tk.StringVar(self.root, name), 8),
            ('Centre_i:', tk.StringVar(self.root, cen_i), 6),
            ('Centre_j:', tk.StringVar(self.root, cen_j), 6),
            ('Width_i', tk.IntVar(self.root, wid_i), 6),
            ('Width_j', tk.IntVar(self.root, wid_j), 6),
            ('Detector', tk.StringVar(self.root, det_name), 10)
        )
        def copy_roi():
            self.add_roi(*(var[1].get() for var in tkvars))

        self.roi_values.append(tkvars)
        ln = ttk.Frame(self.roi_table, borderwidth=2)
        ln.pack(side=tk.TOP, fill=tk.X)
        self.roi_lines.append((True, ln))
        for label, tkvar, width in tkvars:
            ttk.Label(ln, text=label).pack(side=tk.LEFT, padx=2)
            ttk.Entry(ln, textvariable=tkvar, width=width).pack(side=tk.LEFT, padx=2)
        ttk.Button(ln, text='X', command=lambda index=len(self.roi_lines)-1: self.remove_roi(index)).pack(side=tk.LEFT, padx=2)
        ttk.Button(ln, text='Copy', command=copy_roi).pack(side=tk.LEFT, padx=2)

    def remove_roi(self, index: int):
        check, frame = self.roi_lines[index]
        frame.pack_forget()
        self.roi_lines[index] = (False, frame)

    def create_config_rois(self):
        rois = [
            tuple(var[1].get() for var in roi_values)
            for index, roi_values in enumerate(self.roi_values)
            if self.roi_lines[index][0] is True and roi_values[0][1].get()
        ]
        return rois

    def update_config(self):
        self.config[C.roi] = self.create_config_rois()

    def close(self):
        self.update_config()
        if self.close_func:
            self.close_func()
        else:
            self.root.destroy()




"""
scan fit manager for Scan objects
"""

import numpy as np
from lmfit.model import ModelResult, Model, Parameters
from lmfit.models import LinearModel

from mmg_toolbox.nexus.nexus_scan import NexusScan
from .functions import peak_ratio, find_peaks
from .results import FitResults
from .fit_functions import peakfit, multipeakfit, generate_model, generate_model_script

__all__ = ['ScanFitManager']


class ScanFitManager:
    """
    ScanFitManager
     Holds several functions for automatically fitting scan data

    fit = ScanFitManager(scan)
    fit.peak_ratio(yaxis)  # calculates peak power
    fit.find_peaks(xaxis, yaxis)  # automated peak finding routine
    fit.fit(xaxis, yaxis)  # estimate & fit data against a peak profile model using lmfit
    fit.multi_peak_fit(xaxis, yaxis)  # find peaks & fit multiprofile model using lmfit
    fit.model_fit(xaxis, yaxis, model, pars)  # fit supplied model against data
    fit.fit_results()  # return lmfit.ModelResult for last fit
    fit.fit_values()  # return dict of fit values for last fit
    fit.fit_report()  # return str of fit report
    fit.plot()  # plot last lmfit results
    * xaxis, yaxis are str names of arrays in the scan namespace

    :param scan: babelscan.Scan
    """

    def __init__(self, scan: NexusScan):
        self.scan = scan

    def __call__(self, *args, **kwargs) -> FitResults:
        """Calls ScanFitManager.fit(...)"""
        return self.multi_peak_fit(*args, **kwargs)

    def __str__(self):
        return self.fit_report()

    def peak_ratio(self, yaxis: str = 'signal') -> float:
        """
        Return the ratio signal / error for given dataset
        From Blessing, J. Appl. Cryst. (1997). 30, 421-426 Equ: (1) + (6)
          peak_ratio = (sum((y-bkg)/dy^2)/sum(1/dy^2)) / sqrt(i/sum(1/dy^2))
        :param yaxis: str name or address of array to plot on y axis
        :return: float ratio signal / err
        """
        values = self.scan.eval(yaxis)
        errors = np.sqrt(values + 0.1)
        return peak_ratio(values, errors)

    def find_peaks(self, xaxis: str = 'axes', yaxis: str = 'signal', min_peak_power: float | None = None,
                   peak_distance_idx: int = 6) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Find peak shaps in linear-spaced 1d arrays with poisson like numerical values

        E.G.
          centres, index, power = self.find_peaks(xaxis, yaxis, min_peak_power=None, peak_distance_idx=10)

        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
        :param peak_distance_idx: int, group adjacent maxima if closer in index than this
        :return centres: array(m) of peak centers in x, equiv. to xdata[index]
        :return index: array(m) of indexes in y of peaks that satisfy conditions
        :return power: array(m) of estimated power of each peak
        """
        data = self.scan.get_plot_data(xaxis, yaxis)
        xdata, ydata, errors = (data[k] for k in ('xdata', 'ydata', 'yerror'))

        index, power = find_peaks(ydata, errors, min_peak_power, peak_distance_idx)
        return xdata[index], index, power

    def fit(self, xaxis: str = 'axes', yaxis: str = 'signal', model: str = 'Gaussian',
            background: str = 'slope', initial_parameters: dict | None = None,
            fix_parameters: dict | None = None, method: str = 'leastsq',
            print_result: bool = False, plot_result: bool = False) -> FitResults:
        """
        Fit x,y data to a peak model using lmfit

        E.G.:
          res = self.fit('axes', 'signal', model='Gauss')
          print(res)
          res.plot()
          val1 = res.p1_amplitude
          val2 = res.p2_amplitude

        Peak Models:
         Choice of peak model: 'Gaussian', 'Lorentzian', 'Voight', 'PseudoVoight'
        Background Models:
         Choice of background model: 'flat', 'slope', 'exponential'

        Peak Parameters (%d=number of peak):
         'amplitude', 'center', 'sigma', pvoight only: 'fraction'
         output only: 'fwhm', 'height'
        Background parameters:
         'bkg_slope', 'bkg_intercept', or for exponential: 'bkg_amplitude', 'bkg_decay'
         output only: 'background'
        Uncertainties (errors):
         'stderr_PARAMETER', e.g. 'stderr_amplitude'

        Provide initial guess:
          res = self.fit(x, y, model='Voight', initial_parameters={'p1_center':1.23})

        Fix parameter:
          res = self.fit(x, y, model='gauss', fix_parameters={'p1_sigma': fwhm/2.3548200})

        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param model: str, specify the peak model 'Gaussian','Lorentzian','Voight'
        :param background: str, specify the background model: 'slope', 'exponential'
        :param initial_parameters: None or dict of initial values for parameters
        :param fix_parameters: None or dict of parameters to fix at positions
        :param method: str method name, from lmfit fitting methods
        :param print_result: if True, prints the fit results using fit.fit_report()
        :param plot_result: if True, plots the results using fit.plot()
        :return: FitResult object
        """
        data = self.scan.get_plot_data(xaxis, yaxis)
        xdata, ydata, errors, xname, yname = (data[k] for k in ('xdata', 'ydata', 'yerror', 'xlabel', 'ylabel'))

        # lmfit
        res = peakfit(xdata, ydata, errors, model=model, background=background,
                      initial_parameters=initial_parameters, fix_parameters=fix_parameters, method=method)

        output = res.results()
        output.update({
            'lmfit': res.res,
            'fit_result': res.res,
            'fitobj': res,
            'fit': res.res.best_fit,
            f"fit_{yname}": res.res.best_fit,
        })
        self.scan.map.add_local(**output)

        if print_result:
            print(self.scan.title())
            print(res)
        if plot_result:
            res.plot(title=self.scan.title())
        return res

    def multi_peak_fit(self, xaxis: str = 'axes', yaxis: str = 'signal',
                       npeaks: int | None = None, min_peak_power: int | None = None, peak_distance_idx: int = 6,
                       model: str = 'Gaussian', background: str = 'slope',
                       initial_parameters: dict | None = None, fix_parameters: dict | None = None,
                       method: str = 'leastsq', print_result: bool = False, plot_result: bool = False
                       ) -> FitResults:
        """
        Fit x,y data to a peak model using lmfit

        E.G.:
          res = self.multi_peak_fit('axes', 'signal', npeaks=2, model='Gauss')
          print(res)
          res.plot()
          val1 = res.p1_amplitude
          val2 = res.p2_amplitude

        Peak centers:
         Will attempt a fit using 'npeaks' peaks, with centers defined by defalult by the find_peaks function
          if 'npeaks' is None, the number of peaks will be found by find_peaks()
          if 'npeaks' is greater than the number of peaks found by find_peaks, initial peak centers are evenly
          distrubuted along xdata.

        Peak Models:
         Choice of peak model: 'Gaussian', 'Lorentzian', 'Voight','PseudoVoight'
        Background Models:
         Choice of background model: 'flat', 'slope', 'exponential'

        Peak Parameters (%d=number of peak):
         'p%d_amplitude', 'p%d_center', 'p%d_sigma', pvoight only: 'p%d_fraction'
         output only: 'p%d_fwhm', 'p%d_height'
        Background parameters:
         'bkg_slope', 'bkg_intercept', or for exponential: 'bkg_amplitude', 'bkg_decay'
        Total parameters (always available, output only - sum/averages of all peaks):
         'amplitude', 'center', 'sigma', 'fwhm', 'height', 'background'
        Uncertainties (errors):
         'stderr_PARAMETER', e.g. 'stderr_amplitude'

        Provide initial guess:
          res = self.multi_peak_fit(x, y, model='Voight', initial_parameters={'p1_center':1.23})

        Fix parameter:
          res = self.multi_peak_fit(x, y, model='gauss', fix_parameters={'p1_sigma': fwhm/2.3548200})

        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
        :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
        :param peak_distance_idx: int, group adjacent maxima if closer in index than this
        :param model: str, specify the peak model 'Gaussian','Lorentzian','Voight'
        :param background: str, specify the background model: 'slope', 'exponential'
        :param initial_parameters: None or dict of initial values for parameters
        :param fix_parameters: None or dict of parameters to fix at positions
        :param method: str method name, from lmfit fitting methods
        :param print_result: if True, prints the fit results using fit.fit_report()
        :param plot_result: if True, plots the results using fit.plot()
        :return: FitResults object
        """
        data = self.scan.get_plot_data(xaxis, yaxis)
        xdata, ydata, errors, xname, yname = (data[k] for k in ('xdata', 'ydata', 'yerror', 'xlabel', 'ylabel'))

        # lmfit
        res = multipeakfit(xdata, ydata, errors, npeaks=npeaks, min_peak_power=min_peak_power,
                           peak_distance_idx=peak_distance_idx, model=model, background=background,
                           initial_parameters=initial_parameters, fix_parameters=fix_parameters, method=method)

        output = res.results()
        output.update({
            'lmfit': res.res,
            'fit_result': res.res,
            'fitobj': res,
            'fit': res.res.best_fit,
            f"fit_{yname}": res.res.best_fit,
        })
        self.scan.map.add_local(**output)

        if print_result:
            print(self.scan.title())
            print(res)
        if plot_result:
            res.plot(title=self.scan.title())
        return res

    def modelfit(self, xaxis: str = 'axis', yaxis: str = 'signal', model: Model | None = None,
                 pars: Parameters | None = None, method: str = 'leastsq',
                 print_result: bool = False, plot_result: bool = False) -> ModelResult:
        """
        Fit data from scan against lmfit model

        Example:
            from lmfit.models import GaussianModel, LinearModel
            mod = GaussainModel(prefix='p1_') + LinearModel(prefix='bkg_')
            pars = mod.make_params()
            pars['p1_center'].set(value=np.mean(x), min=x.min(), max=x.max())
            res = scan.fit.modelfit('axis', 'signal', mod, pars)
            print(res.fit_report())
            res.plot()
            area = res.params['p1_amplitude'].value
            err = res.params['p1_amplitude'].stderr

        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param model: lmfit.Model - object defining combination of models
        :param pars: lmfit.Parameters - object defining model parameters
        :param method: str name of fitting method to use
        :param print_result: bool, if True, print results.fit_report()
        :param plot_result: bool, if True, generate results.plot()
        :return: lmfit fit results
        """
        data = self.scan.get_plot_data(xaxis, yaxis)
        xdata, ydata, yerror, xname, yname = (data.get(k) for k in ['x', 'y', 'yerror', 'xlabel', 'ylabel'])

        # weights
        if yerror is None or np.all(np.abs(yerror) < 0.001):
            weights = None
        else:
            weights = 1 / np.square(yerror, dtype=float)
            weights = np.nan_to_num(weights)

        # Default model, pars
        if model is None:
            model = LinearModel()
        if pars is None:
            pars = model.guess(ydata, x=xdata)

        # lmfit
        res = model.fit(ydata, pars, x=xdata, weights=weights, method=method)

        fit_dict = {
            'lmfit': res,
            'fit_result': res,
            'fit': res.best_fit,
            f"fit_{yname}": res.best_fit,
        }
        for pname, param in res.params.items():
            ename = 'stderr_' + pname
            fit_dict[pname] = param.value
            fit_dict[ename] = param.stderr

        # Add peak components
        comps = res.eval_components(x=xdata)
        for component in comps.keys():
            fit_dict[f"{component}fit"] = comps[component]
        self.scan.map.add_local(**fit_dict)

        if print_result:
            print(self.scan.title())
            print(res.fit_report())
        if plot_result:
            fig = res.plot(xlabel=xname, ylabel=yname)
            try:
                fig, grid = fig  # Old version of LMFit
            except TypeError:
                pass
            ax1, ax2 = fig.axes
            ax1.set_title(self.scan.title(), wrap=True)
        return res

    def gen_model(self, xaxis: str = 'axes', yaxis: str = 'signal',
                  npeaks: int | None = None, min_peak_power: float | None = None,
                  peak_distance_idx: int = 6, model: str = 'Gaussian', background: str = 'slope',
                  initial_parameters: dict | None = None, fix_parameters: dict | None = None
                  ) -> tuple[Model, Parameters]:
        """
        Generate lmfit model and parameters

        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
        :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
        :param peak_distance_idx: int, group adjacent maxima if closer in index than this
        :param model: str, specify the peak model 'Gaussian','Lorentzian','Voight'
        :param background: str, specify the background model: 'slope', 'exponential'
        :param initial_parameters: None or dict of initial values for parameters
        :param fix_parameters: None or dict of parameters to fix at positions
        :returns: model, pars
        """
        data = self.scan.get_plot_data(xaxis, yaxis)
        xdata, ydata = (data[k] for k in ['x', 'y'])
        mod, pars = generate_model(xdata, ydata,
                                   npeaks=npeaks, min_peak_power=min_peak_power, peak_distance_idx=peak_distance_idx,
                                   model=model, background=background,
                                   initial_parameters=initial_parameters, fix_parameters=fix_parameters)
        return mod, pars

    def gen_model_script(self, xaxis: str = 'axes', yaxis: str = 'signal',
                         npeaks: int | None = None, min_peak_power: float | None = None, peak_distance_idx: int = 6,
                         model: str = 'Gaussian', background: str = 'slope',
                         initial_parameters: dict | None = None, fix_parameters: dict | None = None,
                         only_lmfit: bool = False) -> str:
        """
        Generate script string of fit process
        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
        :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
        :param peak_distance_idx: int, group adjacent maxima if closer in index than this
        :param model: str, specify the peak model 'Gaussian','Lorentzian','Voight'
        :param background: str, specify the background model: 'slope', 'exponential'
        :param initial_parameters: None or dict of initial values for parameters
        :param fix_parameters: None or dict of parameters to fix at positions
        :param only_lmfit: if True, only include imports for lmfit
        :return: str
        """
        data = self.scan.get_plot_data(xaxis, yaxis)
        xdata, ydata, yerror = (data.get(k) for k in ['x', 'y', 'yerror'])
        out = generate_model_script(xdata, ydata, yerror,
                                    npeaks=npeaks, min_peak_power=min_peak_power,
                                    peak_distance_idx=peak_distance_idx,
                                    model=model, background=background,
                                    initial_parameters=initial_parameters, fix_parameters=fix_parameters,
                                    only_lmfit=only_lmfit)
        return out

    def gen_lmfit_script(self, xaxis: str = 'axes', yaxis: str = 'signal',
                         npeaks: int | None = None, min_peak_power: float | None = None, peak_distance_idx: int = 6,
                         model: str = 'Gaussian', background: str = 'slope',
                         initial_parameters: dict | None = None, fix_parameters: dict | None = None, ):
        """
        Generate script string of fit process, using only lmfit

        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
        :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
        :param peak_distance_idx: int, group adjacent maxima if closer in index than this
        :param model: str, specify the peak model 'Gaussian','Lorentzian','Voight'
        :param background: str, specify the background model: 'slope', 'exponential'
        :param initial_parameters: None or dict of initial values for parameters
        :param fix_parameters: None or dict of parameters to fix at positions
        :return: str
        """
        data = self.scan.get_plot_data(xaxis, yaxis)
        xdata, ydata, yerror = (data.get(k) for k in ['x', 'y', 'yerror'])
        out = generate_model_script(xdata, ydata, yerror,
                                    npeaks=npeaks, min_peak_power=min_peak_power,
                                    peak_distance_idx=peak_distance_idx,
                                    model=model, background=background,
                                    initial_parameters=initial_parameters, fix_parameters=fix_parameters,
                                    only_lmfit=True)
        return out

    def fit_parameter(self, parameter_name: str = 'amplitude') -> tuple[float, float]:
        """
        Returns parameter, error from the last run fit
        :param parameter_name: str, name from last fit e.g. 'amplitude', 'center', 'fwhm', 'background'
        :returns:  value, error
        """
        fitobj = self.fit_result()
        return fitobj.get_value(parameter_name)

    def fit_result(self) -> FitResults:
        """
        Returns FitResults object from last fit
        :return: PeakResults obect
        """
        return self.scan('fitobj')

    def fit_report(self) -> str:
        """Return str results of last fit"""
        fitobj = self.fit_result()
        return str(fitobj)

    def plot(self):
        """Plot fit results"""
        fitobj = self.fit_result()
        return fitobj.plot(title=self.scan.title())




"""
lmfit fit wrappers
"""

import numpy as np
from lmfit.model import ModelResult, Model, Parameters

from .functions import gen_weights, find_peaks
from .models import get_default_model, get_peak_model, get_background_model
from .results import FitResults, peak_results_str, peak_results_plot

__all__ = ['modelfit', 'peakfit', 'peak2dfit', 'generate_model', 'generate_model_script', 'multipeakfit']


def modelfit(xvals: np.ndarray, yvals: np.ndarray, yerrors: np.ndarray | None = None,
             model=None | Model, initial_parameters: dict | None = None,
             fix_parameters: dict | None = None, method: str = 'leastsq',
             print_result: bool = False, plot_result: bool = False) -> ModelResult:
    """
    Fit x,y data to a model from lmfit
    E.G.:
      res = modelfit(x, y, model='Gauss')
      print(res.fit_report())
      res.plot()
      val = res.params['amplitude'].value
      err = res.params['amplitude'].stderr

    Model:
     from lmfit import models
     model1 = model.GaussianModel()
     model2 = model.LinearModel()
     model = model1 + model2
     res = model.fit(y, x=x)

    Provide initial guess:
      res = modelfit(x, y, model=VoightModel(), initial_parameters={'center':1.23})

    Fix parameter:
      res = modelfit(x, y, model=VoightModel(), fix_parameters={'sigma': fwhm/2.3548200})

    :param xvals: array(n) position data
    :param yvals: array(n) intensity data
    :param yerrors: None or array(n) - error data to pass to fitting function as weights: 1/errors^2
    :param model: lmfit.Model
    :param initial_parameters: None or dict of initial values for parameters
    :param fix_parameters: None or dict of parameters to fix at positions
    :param method: str method name, from lmfit fitting methods
    :param print_result: if True, prints the fit results using fit.fit_report()
    :param plot_result: if True, plots the results using fit.plot()
    :return: lmfit.model.ModelResult < fit results object
    """

    xvals = np.asarray(xvals, dtype=float).reshape(-1)
    yvals = np.asarray(yvals, dtype=float).reshape(-1)
    weights = gen_weights(yerrors)

    if initial_parameters is None:
        initial_parameters = {}
    if fix_parameters is None:
        fix_parameters = {}

    if model is None:
        model = get_default_model()

    pars = model.make_params()

    # user input parameters
    for ipar, ival in initial_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=True)
    for ipar, ival in fix_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=False)

    res = model.fit(yvals, pars, x=xvals, weights=weights, method=method)

    if print_result:
        print(res.fit_report())
    if plot_result:
        res.plot()
    return res


def peakfit(xvals: np.ndarray, yvals: np.ndarray, yerrors: np.ndarray | None = None,
            model: str = 'Voight', background: str = 'slope',
            initial_parameters: dict | None = None, fix_parameters: dict | None = None,
            method: str = 'leastsq', print_result: bool = False, plot_result: bool = False) -> FitResults:
    """
    Fit x,y data to a peak model using lmfit

    E.G.:
      res = peakfit(x, y, model='Gauss')
      print(res)
      res.plot()
      val = res.params['amplitude'].value
      err = res.params['amplitude'].stderr

    Peak Models:
     Choice of peak model: 'Gaussian', 'Lorentzian', 'Voight',' PseudoVoight'
    Background Models:
     Choice of background model: 'slope', 'exponential'

    Peak Parameters:
     'amplitude', 'center', 'sigma', pvoight only: 'fraction'
     output only: 'fwhm', 'height'
    Background parameters:
     'bkg_slope', 'bkg_intercept', or for exponential: 'bkg_amplitude', 'bkg_decay'

    Provide initial guess:
      res = peakfit(x, y, model='Voight', initial_parameters={'center': 1.23})

    Fix parameter:
      res = peakfit(x, y, model='gauss', fix_parameters={'sigma': fwhm/2.3548200})

    :param xvals: array(n) position data
    :param yvals: array(n) intensity data
    :param yerrors: None or array(n) - error data to pass to fitting function as weights: 1/errors^2
    :param model: str, specify the peak model: 'Gaussian','Lorentzian','Voight'
    :param background: str, specify the background model: 'slope', 'exponential'
    :param initial_parameters: None or dict of initial values for parameters
    :param fix_parameters: None or dict of parameters to fix at positions
    :param method: str method name, from lmfit fitting methods
    :param print_result: if True, prints the fit results using fit.fit_report()
    :param plot_result: if True, plots the results using fit.plot()
    :return: fit result object
    """

    xvals = np.asarray(xvals, dtype=float).reshape(-1)
    yvals = np.asarray(yvals, dtype=float).reshape(-1)
    weights = gen_weights(yerrors)

    if initial_parameters is None:
        initial_parameters = {}
    if fix_parameters is None:
        fix_parameters = {}

    peak_mod = get_peak_model(model)()
    bkg_mod = get_background_model(background)(prefix='bkg_')

    pars = peak_mod.guess(yvals, x=xvals)
    pars += bkg_mod.make_params()
    # pars += bkg_mod.make_params(intercept=np.min(yvals), slope=0)
    # pars['gamma'].set(value=0.7, vary=True, expr='') # don't fix gamma

    # user input parameters
    for ipar, ival in initial_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=True)
    for ipar, ival in fix_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=False)

    mod = peak_mod + bkg_mod
    res = mod.fit(yvals, pars, x=xvals, weights=weights, method=method)

    if print_result:
        print(peak_results_str(res))
    if plot_result:
        peak_results_plot(res)
    return FitResults(res)


def peak2dfit(xdata: np.ndarray, ydata: np.ndarray, image_data: np.ndarray,
              initial_parameters: dict | None = None, fix_parameters: dict | None = None,
              print_result: bool = False, plot_result: bool = False):
    """
    Fit Gaussian Peak in 2D
    *** requires lmfit > 1.0.3 ***
        Not yet finished!
    :param xdata:
    :param ydata:
    :param image_data:
    :param initial_parameters:
    :param fix_parameters:
    :param print_result:
    :param plot_result:
    :return:
    """
    print('Not yet finished...')
    pass


def generate_model(xvals: np.ndarray, yvals: np.ndarray, yerrors: np.ndarray | None = None,
                   npeaks: int | None = None, min_peak_power: float | None = None, peak_distance_idx: int = 6,
                   model: str = 'Gaussian', background: str = 'slope',
                   initial_parameters: dict | None = None, fix_parameters: dict | None = None) -> tuple[
    Model, Parameters]:
    """
    Generate lmfit profile models
    See: https://lmfit.github.io/lmfit-py/builtin_models.html#example-3-fitting-multiple-peaks-and-using-prefixes
    E.G.:
      mod, pars = generate_model(x, y, npeaks=1, model='Gauss', backgroud='slope')

    Peak Search:
     The number of peaks and initial peak centers will be estimated using the find_peaks function. If npeaks is given,
     the largest npeaks will be used initially. 'min_peak_power' and 'peak_distance_idx' can be input to tailor the
     peak search results.
     If the peak search returns < npeaks, fitting parameters will initially choose npeaks equally distributed points

    Peak Models:
     Choice of peak model: 'Gaussian', 'Lorentzian', 'Voight',' PseudoVoight'
    Background Models:
     Choice of background model: 'slope', 'exponential'

    :param xvals: array(n) position data
    :param yvals: array(n) intensity data
    :param yerrors: None or array(n) - error data to pass to fitting function as weights: 1/errors^2
    :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
    :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
    :param peak_distance_idx: int, group adjacent maxima if closer in index than this
    :param model: str or lmfit.Model, specify the peak model 'Gaussian','Lorentzian','Voight'
    :param background: str, specify the background model: 'slope', 'exponential'
    :param initial_parameters: None or dict of initial values for parameters
    :param fix_parameters: None or dict of parameters to fix at positions
    :return: model, parameters
    """
    xvals = np.asarray(xvals, dtype=float).reshape(-1)
    yvals = np.asarray(yvals, dtype=float).reshape(-1)

    # Find peaks
    peak_idx, peak_pow = find_peaks(yvals, yerrors, min_peak_power, peak_distance_idx)
    peak_centers = {'p%d_center' % (n + 1): xvals[peak_idx[n]] for n in range(len(peak_idx))}
    if npeaks is None:
        npeaks = len(peak_centers)

    if initial_parameters is None:
        initial_parameters = {}
    if fix_parameters is None:
        fix_parameters = {}

    peak_mod = get_peak_model(model)
    bkg_mod = get_background_model(background)

    mod = bkg_mod(prefix='bkg_')
    for n in range(npeaks):
        mod += peak_mod(prefix='p%d_' % (n + 1))

    pars = mod.make_params()

    # initial parameters
    min_wid = float(np.mean(np.diff(xvals)))
    max_wid = xvals.max() - xvals.min()
    area = (yvals.max() - yvals.min()) * (3 * min_wid)
    percentile = np.linspace(0, 100, npeaks + 2)
    for n in range(1, npeaks + 1):
        pars['p%d_amplitude' % n].set(value=area / npeaks, min=0)
        pars['p%d_sigma' % n].set(value=3 * min_wid, min=min_wid, max=max_wid)
        pars['p%d_center' % n].set(value=np.percentile(xvals, percentile[n]), min=xvals.min(), max=xvals.max())
    # find_peak centers
    for ipar, ival in peak_centers.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=True)
    # user input parameters
    for ipar, ival in initial_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=True)
    for ipar, ival in fix_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=False)
    return mod, pars


def generate_model_script(xvals: np.ndarray, yvals: np.ndarray, yerrors: np.ndarray = None,
                          npeaks: int | None = None, min_peak_power: float | None = None, peak_distance_idx: int = 6,
                          model: str = 'Gaussian', background: str = 'slope',
                          initial_parameters: dict | None = None, fix_parameters: dict | None = None,
                          only_lmfit: bool = False) -> str:
    """
    Generate script to create lmfit profile models
    E.G.:
      string = generate_mode_stringl(x, y, npeaks=1, model='Gauss', backgroud='slope')

    :param xvals: array(n) position data
    :param yvals: array(n) intensity data
    :param yerrors: None or array(n) - error data to pass to fitting function as weights: 1/errors^2
    :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
    :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
    :param peak_distance_idx: int, group adjacent maxima if closer in index than this
    :param model: str or lmfit.Model, specify the peak model 'Gaussian','Lorentzian','Voight'
    :param background: str, specify the background model: 'slope', 'exponential'
    :param initial_parameters: None or dict of initial values for parameters
    :param fix_parameters: None or dict of parameters to fix at positions
    :param only_lmfit: if True, only include lmfit imports
    :return: str
    """

    data = "xdata = np.array(%s)\n" % list(xvals)
    data += "ydata = np.array(%s)\n" % list(yvals)
    if yerrors is None or np.all(np.abs(yerrors) < 0.001):
        data += 'yerrors = None\n'
        data += 'weights = None\n\n'
    else:
        data += "yerrors = np.array(%s)\n" % list(yerrors)
        data += "yerrors[yerrors < 1] = 1.0\n"
        data += "weights = 1 / yerrors\n\n"

    if initial_parameters is None:
        initial_parameters = {}
    if fix_parameters is None:
        fix_parameters = {}
    params = "initial = %s\nfixed = %s\n" % (initial_parameters, fix_parameters)

    if not only_lmfit:
        out = "import numpy as np\nfrom mmg_toolbox import fitting\n\n"
        out += data
        out += '%s\n' % params
        out += "mod, pars = fitting.generate_model(xdata, ydata, yerrors,\n" \
               "                                   npeaks=%s, min_peak_power=%s, peak_distance_idx=%s,\n" \
               "                                   model='%s', background='%s',\n" \
               "                                   initial_parameters=initial, fix_parameters=fixed)\n" % (
                   npeaks, min_peak_power, peak_distance_idx, model, background
               )
    else:
        # Find peaks
        peak_idx, peak_pow = find_peaks(yvals, yerrors, min_peak_power, peak_distance_idx)
        peak_centers = {'p%d_center' % (n + 1): xvals[peak_idx[n]] for n in range(len(peak_idx))}
        peak_mod = get_peak_model(model)
        bkg_mod = get_background_model(background)
        peak_name = peak_mod.__name__
        bkg_name = bkg_mod.__name__

        out = "import numpy as np\nfrom lmfit import models\n\n"
        out += data
        out += "%speak_centers = %s\n\n" % (params, peak_centers)
        out += "mod = models.%s(prefix='bkg_')\n" % bkg_name
        out += "for n in range(len(peak_centers)):\n    mod += models.%s(prefix='p%%d_' %% (n+1))\n" % peak_name
        out += "pars = mod.make_params()\n\n"
        out += "# initial parameters\n"
        out += "min_wid = np.mean(np.diff(xdata))\n"
        out += "max_wid = xdata.max() - xdata.min()\n"
        out += "area = (ydata.max() - ydata.min()) * (3 * min_wid)\n"
        out += "for n in range(1, len(peak_centers)+1):\n"
        out += "    pars['p%d_amplitude' % n].set(value=area/len(peak_centers), min=0)\n"
        out += "    pars['p%d_sigma' % n].set(value=3*min_wid, min=min_wid, max=max_wid)\n"
        out += "# find_peak centers\n"
        out += "for ipar, ival in peak_centers.items():\n"
        out += "    if ipar in pars:\n"
        out += "        pars[ipar].set(value=ival, vary=True)\n"
        out += "# user input parameters\n"
        out += "for ipar, ival in initial.items():\n"
        out += "    if ipar in pars:\n"
        out += "        pars[ipar].set(value=ival, vary=True)\n"
        out += "for ipar, ival in fixed.items():\n"
        out += "    if ipar in pars:\n"
        out += "        pars[ipar].set(value=ival, vary=False)\n\n"
    out += "# Fit data\n"
    out += "res = mod.fit(ydata, pars, x=xdata, weights=weights, method='leastsqr')\n"
    out += "print(res.fit_report())\n\n"
    out += "fig, grid = res.plot()\n"
    out += "ax1, ax2 = fig.axes\n"
    out += "comps = res.eval_components()\n"
    out += "for component in comps.keys():\n"
    out += "    ax2.plot(xdata, comps[component], mode=component)\n"
    out += "    ax2.legend()\n\n"
    return out


def multipeakfit(xvals: np.ndarray, yvals: np.ndarray, yerrors: np.ndarray | None = None,
                 npeaks: int | None = None, min_peak_power: float | None = None,
                 peak_distance_idx: int | None = 10, model: str = 'Gaussian', background: str = 'slope',
                 initial_parameters: dict | None = None, fix_parameters: dict | None = None,
                 method: str = 'leastsq', remove_peaks: bool = True,
                 print_result: bool = False, plot_result: bool = False) -> FitResults:
    """
    Fit x,y data to a model with multiple peaks using lmfit
    See: https://lmfit.github.io/lmfit-py/builtin_models.html#example-3-fitting-multiple-peaks-and-using-prefixes

    E.G.:
      res = multipeakfit(x, y, npeaks=None, model='Gauss', plot_result=True)
      val = res.params['p1_amplitude'].value
      err = res.params['p1_amplitude'].stderr

    Peak Search:
     The number of peaks and initial peak centers will be estimated using the find_peaks function. If npeaks is given,
     the largest npeaks will be used initially. 'min_peak_power' and 'peak_distance_idx' can be input to tailor the
     peak search results.
     If the peak search returns < npeaks, fitting parameters will initially choose npeaks equally distributed points

    Peak Models:
     Choice of peak model: 'Gaussian', 'Lorentzian', 'Voight',' PseudoVoight'
    Background Models:
     Choice of background model: 'slope', 'exponential'

    Peak Parameters (%d=number of peak):
    Parameters in '.._parameters' dicts and in output results. Each peak (upto npeaks) has a set number of parameters:
     'p%d_amplitude', 'p%d_center', 'p%d_dsigma', pvoight only: 'p%d_fraction'
     output only: 'p%d_fwhm', 'p%d_height'
    Background parameters:
     'bkg_slope', 'bkg_intercept', or for exponential: 'bkg_amplitude', 'bkg_decay'

    Provide initial guess:
      res = multipeakfit(x, y, model='Voight', initial_parameters={'p1_center':1.23})

    Fix parameter:
      res = multipeakfit(x, y, model='gauss', fix_parameters={'p1_sigma': fwhm/2.3548200})

    :param xvals: array(n) position data
    :param yvals: array(n) intensity data
    :param yerrors: None or array(n) - error data to pass to fitting function as weights: 1/errors^2
    :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
    :param min_peak_power: float, only return peaks with power greater than this. If None compares against std(y)
    :param peak_distance_idx: int, group adjacent maxima if closer in index than this
    :param model: str or lmfit.Model, specify the peak model 'Gaussian','Lorentzian','Voight'
    :param background: str, specify the background model: 'slope', 'exponential'
    :param initial_parameters: None or dict of initial values for parameters
    :param fix_parameters: None or dict of parameters to fix at positions
    :param method: str method name, from lmfit fitting methods
    :param remove_peaks: bool, remove peaks consistent with zero and fit again
    :param print_result: if True, prints the fit results using fit.fit_report()
    :param plot_result: if True, plots the results using fit.plot()
    :return: FitResults object
    """
    xvals = np.asarray(xvals, dtype=float).reshape(-1)
    yvals = np.asarray(yvals, dtype=float).reshape(-1)
    weights = gen_weights(yerrors)

    mod, pars = generate_model(xvals, yvals, yerrors,
                               npeaks=npeaks, min_peak_power=min_peak_power, peak_distance_idx=peak_distance_idx,
                               model=model, background=background,
                               initial_parameters=initial_parameters, fix_parameters=fix_parameters)

    # Fit data against model using choosen method
    print(f"Fitting with {len(mod.components) - 1} peaks")
    res = mod.fit(yvals, pars, x=xvals, weights=weights, method=method)
    # Remove peaks consistent with zero
    if remove_peaks and len(mod.components) > 2:
        peak_models = [(m, m.prefix) for m in mod.components if 'bkg' not in m.prefix]
        new_mod = next(m for m in mod.components if 'bkg' in m.prefix)
        for peak_model, prefix in peak_models:
            param = res.params[f"{prefix}amplitude"]
            amplitude = param.value
            stderr = param.stderr or 0
            if amplitude > 2 * stderr:
                new_mod += peak_model

        if len(new_mod.components) < len(mod.components):
            new_pars = new_mod.make_params()
            for par in new_pars:
                res_par = res.params[par]
                new_pars[par].set(value=res_par.value, vary=res_par.vary,
                                  min=res_par.min, max=res_par.max, expr=res_par.expr)
            print(f"Refitting with {len(new_mod.components) - 1} peaks")
            res = new_mod.fit(yvals, new_pars, x=xvals, weights=weights, method=method)

    if print_result:
        print(peak_results_str(res))
    if plot_result:
        peak_results_plot(res)
    return FitResults(res)




"""
mmg_toolbox tests
Test nexus reader
"""
import numpy as np

from mmg_toolbox.nexus.nexus_scan import NexusScan, NexusDataHolder
from mmg_toolbox.nexus.nexus_reader import (read_nexus_file, read_nexus_files, find_scans,
                                            find_similar_scans, find_matching_scans)
from . import only_dls_file_system
from .example_files import DIR

@only_dls_file_system
def test_read_nexus_file():
    f = DIR + '/i16/1109527.nxs'
    scan = read_nexus_file(f)
    assert isinstance(scan, NexusDataHolder)
    assert len(scan.eta_fly) == 61
    assert scan.roi2_sum.max() == 692919
    assert abs(scan.metadata.Tsample-300) < 0.1
    assert scan.scan_number() == 1109527
    assert scan.title() == '#1109527'


@only_dls_file_system
def test_read_nexus_files():
    files = [DIR + f'/i16/cm37262-1/{n}.nxs' for n in range(1032120, 1032130)]
    scans = read_nexus_files(*files)
    assert isinstance(scans[0], NexusScan)


@only_dls_file_system
def test_read_nexus_tiff_files():
    file = DIR + f'/i16/777777.nxs'
    scan, = read_nexus_files(file)
    assert isinstance(scan, NexusScan)
    image = scan.image()
    assert type(image) is np.ndarray
    assert image.shape == (195, 487)


@only_dls_file_system
def test_read_old_nexus_file():
    file = DIR + f'/i16/777777.nxs'
    scan, = read_nexus_files(file)
    data = scan.get_plot_data()
    assert 'x' in data and 'y' in data
    assert data['xlabel'] == 'TimeFromEpoch'
    assert data['ylabel'] == 'sum'
    assert data['x'].size == data['y'].size == 81
    assert len(data['data'].keys()) == 19


@only_dls_file_system
def test_find_scans():
    files = [DIR + f'/i16/cm37262-1/{n}.nxs' for n in range(1032115, 1032135)]
    found_files = find_scans(*files, scan_command='pil3_100k')
    assert len(found_files) == 7
    found_files = find_scans(*files, scan_command='merlin', Ta=(300,0.2))
    assert len(found_files) == 2
    match  = {'start_time.timestamp()': (1705399908, 5)}
    found_files = find_scans(*files, **match)
    assert len(found_files) == 2

